from myqueue.workflow import run


def workflow():
    ph = run(script='phonon.py')
    el = run(script='elph.py', tmax='1h')
    mo = run(script='momentum_matrix.py')
    with ph, el:
        with run(script='supercell_matrix.py', tmax='1h'):
            with run(script='elph_matrix.py'), mo:
                with run(script='raman_intensities.py'):
                    run(function=check)


def check():
    """Read result and make sure it's OK."""
    import numpy as np
    from gpaw.test import findpeak
    ri = np.load('RI_xz_632nm.npy')
    x0, y0 = findpeak(ri[0], ri[1])
    assert np.isclose(x0, 1304.497, atol=0.2)
    assert np.isclose(y0, 5e-5, atol=1e-5)



import os
import shutil

from gpaw.mpi import rank
machine = os.environ.get('MACHINE', 'TEST')
ncores = os.environ.get('NCORES', 8)
if rank == 0:
    os.chdir(machine)
    os.system(
        'python ../memory_bandwidth.py --runs=5 --startcores=' +
        str(ncores))
    shutil.copy('memory_bandwidth_' + machine + '_py.png', '..')



from ase.constraints import FixAtoms
from ase.optimize import QuasiNewton
from ase.calculators.emt import EMT
from ase.build import fcc100, add_adsorbate

from gpaw import GPAW, PW


def aual100(site, height, calc=None):
    slab = fcc100('Al', size=(2, 2, 2))
    slab.center(axis=2, vacuum=3.0)
    add_adsorbate(slab, 'Au', height, site)
    mask = [atom.symbol == 'Al' for atom in slab]
    fixlayer = FixAtoms(mask=mask)
    slab.set_constraint(fixlayer)

    if calc is None:
        calc = GPAW(mode=PW(200), kpts=(2, 2, 1), xc='PBE', txt=site + '.txt',
                    eigensolver='rmm-diis', nbands=40)

    slab.calc = calc
    qn = QuasiNewton(slab, trajectory=site + calc.name + '.traj')
    qn.run(fmax=0.05)

    if calc.name == 'gpaw':
        calc.write(site + '.gpw')

    return slab.get_potential_energy()


e_hollow = aual100('hollow', 1.6)
e_bridge = aual100('bridge', 2.0)
e_ontop = aual100('ontop', 2.4)
assert abs(e_bridge - e_hollow - 0.352) < 0.01
assert abs(e_ontop - e_hollow - 0.711) < 0.01

calc = EMT()
e_hollow = aual100('hollow', 1.6, calc)
e_bridge = aual100('bridge', 2.0, calc)
e_ontop = aual100('ontop', 2.4, calc)
assert abs(e_bridge - e_hollow - 0.401) < 0.01
assert abs(e_ontop - e_hollow - 0.745) < 0.01



from myqueue.workflow import run


def workflow():
    spec_jobs = []

    # Do aug.dzp calculations
    with run(script='basis.py'):
        with run(script='gs.py', cores=4):
            # Simple calculation
            td_jobs = []
            for kick in 'xyz':
                td = run(script='td.py', args=['--kick', kick],
                         cores=4, tmax='1h')
                td_jobs.append(td)
            spec = run(script='spec.py', deps=td_jobs)
            spec_jobs.append(spec)

            # Different origins
            td_jobs = []
            for kick in 'xyz':
                td = run(script='td_origins.py', args=['--kick', kick],
                         cores=4, tmax='1h')
                td_jobs.append(td)
            spec = run(script='spec_origins.py', deps=td_jobs)
            spec_jobs.append(spec)

    # Do dzp calculations
    with run(script='gs.py', cores=4, folder='dzp'):
        td_jobs = []
        for kick in 'xyz':
            td = run(script='td.py', args=['--kick', kick],
                     cores=4, tmax='1h', folder='dzp')
            td_jobs.append(td)
        spec = run(script='spec.py', deps=td_jobs, folder='dzp')
        spec_jobs.append(spec)

    # Plot spectra
    run(script='plot_spectra.py', deps=spec_jobs)
    run(script='plot_spectra_origins.py', deps=spec_jobs)



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Battery Project

## Day 4 - Transport barriers and Voltage profile
"""

# %%
"""
Today you will calculate the energy barriers for transport of Li intercalated in the graphite anode. You will examine how sensitive this barrier is to the interlayer distance in graphite.  You will also examine the energy of intermediate states during the charge/discharge process. This will allow some basic discussion of the voltage profile of the battery.

You will in general be provided less code than yesterday, especially towards the end of this notebook. You will have to use what you have already seen and learned so far.

There will be some natural pauses while you wait for calculations to finish. If you do not finish this entire notebook today, do not despair.
"""

# %%
"""
## Initialize
"""

# %%
# magic: %matplotlib notebook
from ase import Atom
from ase.visualize import view
import matplotlib.pyplot as plt
from ase.io import read, write
from ase.neb import NEB
from ase.optimize import BFGS
from ase.parallel import paropen
from gpaw import GPAW, FermiDirac, Mixer, PW
from ase.constraints import FixAtoms

# %%
"""
## Transport barrier of Li in graphite
"""

# %%
"""
You will now calculate the energy barrier for Li diffusion in the graphite anode. You will do this using the [Nudged Elastic Band (NEB) method](https://wiki.fysik.dtu.dk/ase/ase/neb.html#module-ase.neb)

You can use your work from Day 2, but for simplicity you are advised to load in the initial atomic configuration from file.
"""

# %%
initial = read('NEB_init.traj')

# %%
"""
Visualize the structure.
"""

# %%
view(initial)

# %%
"""
You will now make a final structure, where the Li atom has been moved to a neighbouring equivalent site. The [`get_positions`](https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=get_positions#ase.Atoms.get_positions), [`set_positions`](https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=get_positions#ase.Atoms.set_positions) and [`get_cell`](https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=get_positions#ase.Atoms.get_cell) functions are highly useful for such a task. HINT: Displace the Li atom $\frac{1}{n} (\vec{a}+\vec{b})$
"""

# %%
final = initial.copy()

# %%
# ...
# ...

# teacher
cell = final.get_cell()
pos = final.get_positions()
pos[6] = pos[6] + cell[1] / 3 + cell[0] / 3
final.set_positions(pos)

# %%
"""
Visualize that you have made the final strcuture correctly.
"""

# %%
view(final)

# %%
"""
Make a band consisting of 7 images including the initial and final.
"""

# %%
images = [initial]
images += [initial.copy() for i in range(5)]  # These will become the minimum energy path images.
images += [final]

# %%
"""
It this point `images` consist of 6 copies of `initial` and one entry of `final`. Use the `NEB` method to create an initial guess for the minimum energy path (MEP). In the cell below a simple interpolation between the `initial` and `final` image is used as initial guess.
"""

# %%
neb = NEB(images)
neb.interpolate()

# %%
"""
Visualize the NEB images.
"""

# %%
view(images)

# %%
"""
It turns out, that while running the NEB calculation, the largest amount of resources will be spend translating the carbon layer without any noticeable buckling. You will thus [constrain](https://wiki.fysik.dtu.dk/ase/ase/constraints.html#constraints) the positions of the carbon atoms to save computational time.

Each image in the NEB requires a unique calculator.

This very simple case is highly symmetric. To better illustrate how the NEB method works, the symmetry is broken using the [rattle](https://wiki.fysik.dtu.dk/ase/ase/atoms.html#ase.Atoms.rattle) function.
"""

# %%
for image in images[0:7]:
    calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='LDA', txt=None, symmetry={'point_group': False})
    image.set_calculator(calc)
    image.set_constraint(FixAtoms(mask=[atom.symbol == 'C' for atom in image]))

images[3].rattle(stdev=0.05, seed=42)


# %%
"""
Start by calculating the energy and forces of the first (`initial`) and last (`final`) images as this is not done during the actual NEB calculation.

Note, that this can take a while if you opt to do it inside the notebook.
"""

# %%
images[0].get_potential_energy()
images[0].get_forces()
images[6].get_potential_energy()
images[6].get_forces()


# %%
"""
You can run the NEB calculation by running an optimization on the NEB object the same way you would on an atoms object. Note the `fmax` is larger for this tutorial example than you would normally use.
"""

# %%
optimizer = BFGS(neb, trajectory='neb.traj', logfile='neb.log' )
optimizer.run(fmax=0.10)


# %%
"""
Submit the calculation to the HPC cluster. Do this by first building a complete script in the cell below using the cells above (minus the `view()` commands). Make sure the cell runs and then interrupt the kernel.
"""

# %%
#from ase.io import read, write
#from ase.neb import NEB
#from ase.optimize import BFGS
#from ase.parallel import paropen
#from gpaw import GPAW, FermiDirac, Mixer, PW
#from ase.constraints import FixAtoms

# initial = read('NEB_init.traj')

# final = ...

# ...
# ...

# optimizer.run(fmax=0.10)

# teacher
from ase.io import read, write
from ase.neb import NEB
from ase.optimize import BFGS
from ase.parallel import paropen
from gpaw import GPAW, FermiDirac, Mixer, PW
from ase.constraints import FixAtoms

initial=read('NEB_init.traj')

final=initial.copy()
cell=final.get_cell()
pos=final.get_positions()
pos[6]=pos[6]+cell[1]/3.+cell[0]/3.
final.set_positions(pos)

images = [initial]
images += [initial.copy() for i in range(5)]  #These will become the minimum energy path images.
images += [final]

neb = NEB(images)
neb.interpolate()

for image in images[0:7]:
    calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='LDA', symmetry={'point_group': False})
    image.calc = calc
    image.set_constraint(FixAtoms(mask=[atom.symbol == 'C' for atom in image]))

images[3].rattle(stdev=0.05, seed=42)

images[0].get_potential_energy()
images[0].get_forces()
images[6].get_potential_energy()
images[6].get_forces()

optimizer = BFGS(neb, trajectory = 'neb.traj', logfile = 'neb.log' )
optimizer.run(fmax = 0.10)

# %%
"""
You can use the cell below to submit the calculation in the same way as on earlier days.
"""

# %%
# magic: !mq submit NEB.py -R 8:1h  # submits the calculation to 8 cores, 1 hour

# %%
"""
Run the below cell to examine the status of your calculation.
"""

# %%
# magic: !mq ls

# %%
"""
You can run the cells below to open the error log and output of the calculation in a new window. This can be done while the calculation is running.
"""

# %%
# magic: !cat "$(ls -t NEB.py.*err | head -1)"

# %%
# magic: !cat "$(ls -t NEB.py.*out | head -1)"

# %%
"""
The optimiziation progress can be seen by running the below cell.
"""

# %%
# magic: !cat neb.log

# %%
"""
You can move on while you wait for the calculation to finish.

Once the maximum force (`fmax`) in the log is below 0.1, the calculation is finished.
Load in the full trajectory.
"""

# %%
full_neb = read('neb.traj@:')

# %%
"""
You will use the `ase gui` to inspect the result. The below line reads in the last 7 images in the file. In this case the MEP images.
"""

# %%
# magic: !ase gui neb.traj@-7:

# %%
"""
In the GUI use `Tools` $\rightarrow$ `NEB`.

Now inspect how the TS image has developed.
"""

# %%
# magic: !ase gui neb.traj@3::7

# %%
"""
For more complicated MEP's, use the [climbing image method](https://wiki.fysik.dtu.dk/ase/ase/neb.html?highlight=neb#climbing-image) to determine the transition state. Why is it not required here?
"""

# %%
"""
## Bonus

You will now study the influence of changing the interlayer graphite distance on the energy barrier. Due to the high degree of symmetry, this can be done easily in this case. Load in the initial state (IS) and transition state (TS) images from the converged MEP.
"""

# %%
IS_image = images[0]
TS_image = images[3]

# %%
"""
Now calculate the energy of the initial state (IS) image and the transition state (TS) image using [`get_potential_energy()`](https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=get_potential_energy#ase.Atoms.get_potential_energy)
"""

# %%
epot_IS = IS_image.get_potential_energy()
#epot_TS= ...

# teacher
epot_TS = TS_image.get_potential_energy()

# %%
barrier = epot_TS - epot_IS
print('Energy barrier:', barrier)

# %%
"""
Why does this not fully align with what you found before? New reduce the graphite layer distance by change the the size of the unit cell in the *z* direction by 3 %.
"""

# %%
cell = IS_image.get_cell()
IS_image97 = IS_image.copy()
IS_image97.set_cell([cell[0], cell[1], cell[2] * 0.97], scale_atoms=True)
TS_image97 = TS_image.copy()
TS_image97.set_cell([cell[0], cell[1], cell[2] * 0.97], scale_atoms=True)

# %%
"""
Use the same calculator object as you did above and calculate the potential energy of the compressed initial and final state.
"""

# %%
# calc = ...
# ...

# teacher
calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='LDA', symmetry={'point_group': False})
TS_image97.calc = calc
calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='LDA', symmetry={'point_group': False})
IS_image97.calc = calc

# %%
"""
Now calculate the energy of the compressed IS and TS.
"""

# %%
# epot_TS97 = ...

# teacher
epot_TS97 = TS_image97.get_potential_energy()
epot_IS97 = IS_image97.get_potential_energy()

# %%
"""
What is the energy barrier now?
"""

# %%
# barrier97 = ...
# print('Energy barrier:', barrier97)

# teacher
barrier97=epot_TS97-epot_IS97
print("Energy barrier:", barrier97)

# %%
"""
Now repeat the procedure but expanding the intergraphite distance by 3 %.
"""

# %%
# IS_image103 = IS_image.copy()
# IS_image103.set_cell(...

# calc ...


# epot_TS103 = ...
# ...

# teacher
IS_image103=IS_image.copy()
IS_image103.set_cell([cell[0],cell[1],cell[2]*1.03], scale_atoms=True)
TS_image103=TS_image.copy()
TS_image103.set_cell([cell[0],cell[1],cell[2]*1.03], scale_atoms=True)

calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='LDA', symmetry={'point_group': False})
TS_image103.calc = calc
calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='LDA', symmetry={'point_group': False})
IS_image103.calc = calc

epot_TS103=TS_image103.get_potential_energy()
epot_IS103=IS_image103.get_potential_energy()

# %%
"""
What is the energy barrier now?
"""

# %%
# barrier103 = ...
# print('Energy barrier:', barrier103)

# teacher
barrier103 = epot_TS103 - epot_IS103
print('Energy barrier:', barrier103)

# %%
"""
## FePO$_4$ with one Li
"""

# %%
"""
You will now calculate the energy gain of adding a single Li atom into the FePO$_4$ cell you made on Day 3. This corresponds to a charge of 25 %. You can compare this energy to the equilibrium potential.

Load in the FePO$_4$ structure you wrote to file on in a previous exercise and add Li. Assume that the cell dimension remain unchanged.
"""

# %%
#fepo4=read('fepo4.traj')
#fepo4_1li=fepo4.copy()

# teacher
from ase import Atoms
fepo4=Atoms('FeFeFeFeOOOOOOOOOOOOOOOOPPPP',
             positions=[[2.73015081, 1.46880951, 4.56541172],
                [2.23941067, 4.40642872, 2.14957739],
                [7.20997230, 4.40642925, 0.26615813],
                [7.70070740, 1.46880983, 2.68199421],
                [1.16033403, 1.46881052, 3.40240205],
                [3.80867172, 4.40642951, 0.98654342],
                [8.77981469, 4.40642875, 1.42923946],
                [6.13142032, 1.46881092, 3.84509827],
                [4.37288562, 1.46880982, 0.81812712],
                [0.59764596, 4.40643021, 3.23442747],
                [5.56702590, 4.40642886, 4.01346264],
                [9.34268360, 1.46880929, 1.59716233],
                [1.64001691, 0.26061277, 1.17298291],
                [3.32931769, 5.61463705, 3.58882629],
                [8.30013707, 3.19826250, 3.65857000],
                [6.61076951, 2.67698811, 1.24272700],
                [8.30013642, 5.61459688, 3.65856912],
                [6.61076982, 0.26063178, 1.24272567],
                [1.64001666, 2.67700652, 1.17298270],
                [3.32931675, 3.19822249, 3.58882660],
                [0.90585688, 1.46880966, 1.89272372],
                [4.06363530, 4.40642949, 4.30853266],
                [9.03398503, 4.40642957, 2.93877879],
                [5.87676435, 1.46881009, 0.52297232]
                        ],
             cell=[9.94012, 5.87524, 4.83157],
             pbc=[1, 1, 1])

for atom in fepo4:
    if atom.symbol == 'Fe':
        atom.magmom = 5.0

fepo4_1li = fepo4.copy()

# %%
# fepo4_1li.append(...)

# teacher
fepo4_1li.append('Li')

# %%
"""
Visualize the structure you made.
"""

# %%
view(fepo4_1li)

# %%
"""
Adjust the total magnetic moment of the cell such that it is 19.
"""

# %%
for atom in fepo4_1li:
    if atom.symbol == 'Fe':
        atom.magmom = 4.75

print(sum(fepo4_1li.get_initial_magnetic_moments()))

# %%
"""
Write your atoms object to file giving it the name `fepo4_1li.traj`.
"""

# %%
write('fepo4_1li.traj', fepo4_1li)

# %%
"""
Make a full script in the cell below similar to those you made yesterday. Make sure the cell runs before interupting the notebook kernel.
"""

# %%
# %%writefile 'fepo4_1li.py'
#from ase.parallel import paropen
#from ase.io import read, write
#from ase.dft.bee import BEEFEnsemble
#from gpaw import GPAW, FermiDirac, Mixer, PW

# Read in the structure you made and wrote to file above
fepo4_1li = read('fepo4_1li.traj')

#...
#...

# write('fepo4_1li_out.traj', fepo4_1li)

# ens = BEEFEnsemble(calc)
# with paropen('ensemble_fepo4_1li.dat', 'a') as result:
#     for e in dE:
#         print(e, file=result)

# teacher
from ase.parallel import paropen
from ase.io import read
from ase.dft.bee import BEEFEnsemble
from gpaw import GPAW, FermiDirac, Mixer, PW

#Read in the structure you made and wrote to file above
fepo4_1li=read('fepo4_1li.traj')

params_GPAW = {}
params_GPAW['mode']        = PW(500)                     #The used plane wave energy cutoff
params_GPAW['nbands']      = -40                           #The number on empty bands had the system been spin-paired
params_GPAW['kpts']        = {'size':  (2,4,5),            #The k-point mesh
                              'gamma': True}
params_GPAW['spinpol']     = True                          #Performing spin polarized calculations
params_GPAW['xc']          = 'BEEF-vdW'                    #The used exchange-correlation functional
params_GPAW['occupations'] = FermiDirac(width = 0.1,      #The smearing
                                        fixmagmom = True)  #Total magnetic moment fixed to the initial value
params_GPAW['convergence'] = {'eigenstates': 1.0e-4,       #eV^2 / electron
                              'energy':      2.0e-4,       #eV / electron
                              'density':     1.0e-3,}
params_GPAW['mixer']       = Mixer(0.1, 5, weight=100.0)   #The mixer used during SCF optimization
params_GPAW['setups']      = {'Fe': ':d,4.3'}              #U=4.3 applied to d orbitals

calc = GPAW(**params_GPAW)
fepo4_1li.calc = calc
epot_fepo4_1li_cell=fepo4_1li.get_potential_energy()
print('E_Pot=', epot_fepo4_1li_cell)

write('fepo4_1li_out.traj', fepo4_1li)

ens = BEEFEnsemble(calc)
dE = ens.get_ensemble_energies(2000)
result = paropen('ensemble_fepo4_1li.dat','a')
for i in range(0,len(dE)):
    print(dE[i], file=result)
result.close()

# %%
"""
Submit this calculation to the HPC cluster as you did on exercise day 3.
"""

# %%
# magic: !mq submit fepo4_1li.py -R 8:1h # submits the calculation to 8 cores, 1 hour

# %%
"""
Run the below cell to examine the status of your calculation.
"""

# %%
# magic: !mq ls

# %%
"""
You can run the cells below to open the error log and output of the calculation in a new window. This can be done while the calculation is running.
"""

# %%
# magic: !cat "$(ls -t fepo4_1li.py.*err | head -1)"

# %%
# magic: !cat "$(ls -t fepo4_1li.py.*out | head -1)"

# %%
"""
You can move on while you wait for the calculation to finish. Once the calculation is finished load in the structure by running the cell below.
"""

# %%
try:
    fepo4_1li=read('fepo4_1li_out.traj')
    print('Calculation finished')
except FileNotFoundError:
    print('Calculation has not yet finished')

# %%
"""
You are now ready to calculate the energy gained by intercalating a single Li ion into the cathode. Start by loading in the relevant reference structures and obtain the potential energies. This should not require any DFT calculations.
"""

# %%
# Loading in files from exercise day 3.
li_metal = read('li_metal.traj')
fepo4 = read('fepo4_out.traj')

epot_li_metal = li_metal.get_potential_energy() / len(li_metal)

# %%
# epot_fepo4 = ...
# ...

# teacher
epot_fepo4=fepo4.get_potential_energy()
epot_fepo4_1li=fepo4_1li.get_potential_energy()

# %%
"""
Calculate the energy of intercalting a single Li in the FePO$_4$ cell. How does this energy compare with the equilibirum potential? What can it tell you about the charge/discharge potential curves?
"""

# %%
# ...
# print(...)

# teacher
li_cost=epot_fepo4_1li-epot_fepo4-epot_li_metal
print(li_cost)

# %%
"""
## Bonus: LiFePO$_4$ with one vacancy
"""

# %%
"""
If time permits, you will now do a similar calculation but this time with LiFePO$_4$ contraining one vacancy. Once again you should assume that the cell dimension remain unchanged compaired to LiFePO$_4$.

There are numerous ways to obtain this structure. You can get inspiration from the way LiFePO$_4$ was made on Exercise day 3, use the [`del` or `pop()` methods](https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=pop#list-methods), or even use the GUI to delete an atom and save the structure afterwards.
"""

# %%
# In this cell you create the vacancy in LiFePO4

# lifepo4_vac = ...

# ...

# teacher
lifepo4_wo_li=read('lifepo4_wo_li.traj')
from numpy import identity
cell=lifepo4_wo_li.get_cell()
xyzcell = identity(3)
lifepo4_wo_li.set_cell(xyzcell, scale_atoms=True)  # Set the unit cell and rescale
#lifepo4_wo_li.append(Atom('Li', (0, 0, 0)))
lifepo4_wo_li.append(Atom('Li', (0, 0.5, 0)))
lifepo4_wo_li.append(Atom('Li', (0.5, 0.5, 0.5)))
lifepo4_wo_li.append(Atom('Li', (0.5, 0, 0.5)))
lifepo4_wo_li.set_cell(cell, scale_atoms=True)
lifepo4_vac=lifepo4_wo_li.copy()

# %%
"""
Visualize the structure
"""

# %%
view(lifepo4_vac)

# %%
"""
Now ensure that the total magnetic moment is equal to 17.
"""

# %%
for atom in fepo4_1li:
    if atom.symbol == 'Fe':
        atom.magmom = 4.25

print(sum(fepo4_1li.get_initial_magnetic_moments()))

# %%
"""
Write your atoms object to file giving it the name `lifepo4_vac.traj`.
"""

# %%
# ...

# teacher
write('lifepo4_vac.traj', lifepo4_vac)

# %%
"""
Make a full script in the cell below similar to that you made above. Make sure the cell runs before interupting the notebook kernel.
"""

# %%
# %%writefile 'lifepo4_vac.py'
# from ase.parallel import paropen
# from ase.io import read, write
# from ase.dft.bee import BEEFEnsemble
# from gpaw import GPAW, FermiDirac, Mixer, PW

# Read in the structure you made and wrote to file above
# lifepo4_vac = read('lifepo4_vac.traj')


# ...

# write('lifepo4_vac_out.traj', lifepo4_vac)

# ens = BEEFEnsemble(calc)
# dE = ens.get_ensemble_energies(2000)
# with paropen('ensemble_lifepo4_vac.dat','a') as results:
#     for e in dE:
#         print(e, file=result)

# teacher
from ase.parallel import paropen
from ase.io import read, write
from ase.dft.bee import BEEFEnsemble
from gpaw import GPAW, FermiDirac, Mixer, PW

# Read in the structure you made and wrote to file above
lifepo4_vac=read('lifepo4_vac.traj')

params_GPAW = {}
params_GPAW['mode']        = PW(500)                     #The used plane wave energy cutoff
params_GPAW['nbands']      = -40                           #The number on empty bands had the system been spin-paired
params_GPAW['kpts']        = {'size':  (2,4,5),            #The k-point mesh
                              'gamma': True}
params_GPAW['spinpol']     = True                          #Performing spin polarized calculations
params_GPAW['xc']          = 'BEEF-vdW'                    #The used exchange-correlation functional
params_GPAW['occupations'] = FermiDirac(width = 0.1,      #The smearing
                                        fixmagmom = True)  #Total magnetic moment fixed to the initial value
params_GPAW['convergence'] = {'eigenstates': 1.0e-4,       #eV^2 / electron
                              'energy':      2.0e-4,       #eV / electron
                              'density':     1.0e-3,}
params_GPAW['mixer']       = Mixer(0.1, 5, weight=100.0)   #The mixer used during SCF optimization
params_GPAW['setups']      = {'Fe': ':d,4.3'}              #U=4.3 applied to d orbitals

calc = GPAW(**params_GPAW)
lifepo4_vac.calc = calc
epot_lifepo4_vac_cell=lifepo4_vac.get_potential_energy()
print('E_Pot=', epot_lifepo4_vac_cell)

write('lifepo4_vac_out.traj', lifepo4_vac)

ens = BEEFEnsemble(calc)
dE = ens.get_ensemble_energies(2000)
result = paropen('ensemble_lifepo4_vac.dat','a')
for i in range(0,len(dE)):
    print(dE[i], file=result)
result.close()

# %%
"""
Once you have made sure the cell runs, submit it to the HPC cluster.
"""

# %%
# magic: !mq submit lifepo4_vac.py -R 8:1h  # submits the calculation to 8 cores, 1 hour

# %%
"""
Once the calculation has finished, load in the trajectory.
"""

# %%
try:
    lifepo4_vac=read('lifepo4_vac_out.traj')
    print('Calculation finished')
except FileNotFoundError:
    print('Calculation has not yet finished')

# %%
"""
Once the calculation has finished you are ready to calculate the energy cost of creating a li vacancy in the fully lithiated LiFePO$_4$. Start by loading in the relevant reference structures and obtain the potential energies. This should not require any calculations.
"""

# %%
# Loading in files from exercise day 3.
li_metal = read('li_metal.traj')   # you should have already read this in above
lifepo4 = read('lifepo4_out.traj')

epot_li_metal = li_metal.get_potential_energy() / len(li_metal)

# %%
# epot_lifepo4 = ...
# ...

# teacher
epot_lifepo4=lifepo4.get_potential_energy()
epot_lifepo4_vac=lifepo4_vac.get_potential_energy()

# %%
# vac_cost = ...
# print(vac_cost)

# teacher
vac_cost=epot_lifepo4_vac-epot_lifepo4+epot_li_metal
print(vac_cost)

# %%
"""
How does this energy compare with the equilibirum potential? What can it tell you about the charge/discharge potential curves?
"""

# %%
"""
## Bonus
Calculate the error estimates of the energy for the added Li atom and vacancy formation using the ensembles.
"""

# %%
# Cell for bonus question

# %%
# Cell for bonus question

# %%
# Cell for bonus question



import sys
from ase import Atoms
from gpaw import GPAW, FermiDirac

# Script to get the total energies of a supercell
# of GaAs with and without a Ga vacancy

a = 5.628  # Lattice parameter
N = int(sys.argv[1])  # NxNxN supercell
q = -3  # Defect charge

formula = 'Ga4As4'

lattice = [[a, 0.0, 0.0],  # work with cubic cell
           [0.0, a, 0.0],
           [0.0, 0.0, a]]

basis = [[0.0, 0.0, 0.0],
         [0.5, 0.5, 0.0],
         [0.0, 0.5, 0.5],
         [0.5, 0.0, 0.5],
         [0.25, 0.25, 0.25],
         [0.75, 0.75, 0.25],
         [0.25, 0.75, 0.75],
         [0.75, 0.25, 0.75]]

GaAs = Atoms(symbols=formula,
             scaled_positions=basis,
             cell=lattice,
             pbc=(1, 1, 1))

GaAsdef = GaAs.repeat((N, N, N))

GaAsdef.pop(0)  # Make the supercell and a Ga vacancy

calc = GPAW(mode='fd',
            kpts={'size': (2, 2, 2), 'gamma': False},
            xc='LDA',
            charge=q,
            occupations=FermiDirac(0.01),
            txt='GaAs{0}{0}{0}.Ga_vac.txt'.format(N))


GaAsdef.calc = calc
Edef = GaAsdef.get_potential_energy()

calc.write('GaAs{0}{0}{0}.Ga_vac_charged.gpw'.format(N))

# Now for the pristine case

GaAspris = GaAs.repeat((N, N, N))
parameters = calc.todict()
parameters['txt'] = 'GaAs{0}{0}{0}.pristine.txt'.format(N)
parameters['charge'] = 0
calc = GPAW(**parameters)

GaAspris.calc = calc
Epris = GaAspris.get_potential_energy()

calc.write('GaAs{0}{0}{0}.pristine.gpw'.format(N))



def workflow():
    from myqueue.workflow import run
    with run(function=init):
        run(script='work_function.py')


def init():
    import runpy
    from pathlib import Path
    source = Path('Al100.py').read_text()
    source = source.replace('k = ...', 'k = 6')
    source = source.replace('N = ...', 'N = 5')
    path = Path('al10065.py')
    path.write_text(source)
    runpy.run_path(str(path))



import ase.db
from ase.build import bulk

from gpaw import GPAW, PW, FermiDirac
from gpaw.response.g0w0 import G0W0


data = {
    'C': ['diamond', 3.567],
    'SiC': ['zincblende', 4.358],
    'Si': ['diamond', 5.431],
    'Ge': ['diamond', 5.658],
    'BN': ['zincblende', 3.616],
    'AlP': ['zincblende', 5.463],
    'AlAs': ['zincblende', 5.661],
    'AlSb': ['zincblende', 6.136],
    'GaN': ['zincblende', 4.50],
    # 'GaN': ['zincblende', ],
    'GaP': ['zincblende', 5.451],
    'GaAs': ['zincblende', 5.654],
    'GaSb': ['zincblende', 6.096],
    'InP': ['zincblende', 5.869],
    'InAs': ['zincblende', 6.058],
    'InSb': ['zincblende', 6.479],
    'MgO': ['zincblende', 4.53],
    'ZnO': ['zincblende', 4.60],
    # 'ZnO': ['zincblende', ],
    'ZnS': ['zincblende', 5.409],
    'ZnSe': ['zincblende', 5.669],
    'ZnTe': ['zincblende', 6.103],
    'CdO': ['zincblende', 5.148],
    'CdS': ['zincblende', 5.818],
    'CdSe': ['zincblende', 6.077],
    'CdTe': ['zincblende', 6.481]}

c = ase.db.connect('gw.db')
tag = 'a4'

for name in data:
    id = c.reserve(name=name, tag=tag)
    if id is None:
        continue

    x, a = data[name]
    atoms = bulk(name, x, a=a)
    atoms.calc = GPAW(mode=PW(600),
                      xc='LDA',
                      parallel={'band': 1},
                      occupations=FermiDirac(0.02),
                      kpts={'size': (4, 4, 4), 'gamma': True},
                      txt='%s.txt' % name)
    atoms.get_potential_energy()
    atoms.calc.diagonalize_full_hamiltonian()
    atoms.calc.write(name, mode='all')
    n = int(atoms.calc.get_number_of_electrons()) // 2
    gw = G0W0(name, name + 'gw',
              nblocks=4,
              kpts=[(0, 0, 0), (0.5, 0.5, 0.5), (0.5, 0.5, 0)],
              ecut=100,
              wstc=True,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.2,
                           'omega2': 10},
              eta=0.2,
              bands=(0, n + 2))
    results = gw.calculate()
    c.write(atoms, name=name, tag=tag, data=results)
    del c[id]



# Takes 30 mins with 8 cores in domain-parallel (memory ~500mb, max 1.5gb)

import numpy as np
from ase import Atom, Atoms
from gpaw import GPAW, MixerDif, mpi
from gpaw.lrtddft import LrTDDFT
from gpaw.pes.dos import DOSPES
from gpaw.pes.tddft import TDDFTPES

atoms = Atoms(
    [Atom('N', (5.915080000001099, 5.584099935483791, 5.517093011559592)),
     Atom('H', (5.915080000000824, 6.534628192918412, 5.154213005717621)),
     Atom('H', (6.739127072810322, 5.112845817889995, 5.151626633318725)),
     Atom('H', (5.091032927190902, 5.112845817890320, 5.151626633319492))])
atoms.center(vacuum=8)

h = 0.15
N_c = np.round(atoms.get_cell().diagonal() / h / 16) * 16

m_c = GPAW(gpts=N_c, nbands=4, mixer=MixerDif(0.1, 5, weight=100.0),
           # convergence={'bands':10},
           parallel={'domain': mpi.size},
           xc='PBE', txt='NH3-m.txt', spinpol=True)

m = atoms.copy()
m.set_initial_magnetic_moments([-1, 1, 1, -1])
m.calc = m_c
m.get_potential_energy()

d_c = GPAW(gpts=N_c, nbands=16, mixer=MixerDif(0.1, 5, weight=100.0),
           convergence={'bands': 10},
           parallel={'domain': mpi.size},
           xc='PBE', txt='NH3-d.txt', spinpol=True)

d = atoms.copy()
d.set_initial_magnetic_moments([-1, 0.5, 0.5, -0.5])
d_c.set(charge=1)
d.calc = d_c
d.get_potential_energy()

istart = 0  # band index of the first occ. band to consider
jend = 15   # band index of the last unocc. band to consider
d_lr = LrTDDFT(d_c, xc='PBE', nspins=2,
               restrict={'istart': istart, 'jend': jend})
d_lr.diagonalize()

pes = TDDFTPES(m_c, d_lr, d_c)
pes.save_folded_pes('NH3-td.dat', folding=None)

pes = DOSPES(m_c, d_c, shift=True)
pes.save_folded_pes('NH3-dos.dat', folding=None)



from __future__ import annotations

from functools import partial
from typing import Callable

import numpy as np
from gpaw import debug
from gpaw.core.arrays import DistributedArrays as DA
from gpaw.core.atom_centered_functions import AtomArrays as AA
from gpaw.core.matrix import Matrix
from gpaw.new.calculation import DFTState
from gpaw.new.eigensolver import Eigensolver
from gpaw.new.hamiltonian import Hamiltonian
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.typing import Array1D, Array2D
from gpaw.utilities.blas import axpy
from gpaw.yml import obj2yaml as o2y
from scipy.linalg import eigh

AAFunc = Callable[[AA, AA], AA]


class Davidson(Eigensolver):
    def __init__(self,
                 nbands,
                 wf_grid,
                 band_comm,
                 preconditioner_factory,
                 niter=2,
                 blocksize=10,
                 converge_bands='occupied',
                 scalapack_parameters=None):
        self.niter = niter
        self.converge_bands = converge_bands

        B = nbands
        domain_comm = wf_grid.comm
        if domain_comm.rank == 0 and band_comm.rank == 0:
            self.H_NN = Matrix(2 * B, 2 * B, wf_grid.dtype)
            self.S_NN = Matrix(2 * B, 2 * B, wf_grid.dtype)
        else:
            self.H_NN = self.S_NN = EmptyMatrix()

        self.M_nn = Matrix(B, B, wf_grid.dtype,
                           dist=(band_comm, band_comm.size))

        self.work_arrays: np.ndarray | None = None

        self.preconditioner = preconditioner_factory(blocksize)

    def __str__(self):
        return o2y(dict(name='Davidson',
                        niter=self.niter,
                        converge_bands=self.converge_bands))

    def iterate(self, state: DFTState, hamiltonian: Hamiltonian) -> float:
        """Iterate on state given fixed hamiltonian.

        Returns
        -------
        float:
            Weighted error of residuals:::

                   ~     ~ ~
              R = (H - ε S)ψ
               n        n   n
        """
        if self.work_arrays is None:
            # First time: allocate work-arrays
            shape = state.ibzwfs.get_max_shape()
            shape = (2, state.ibzwfs.nbands) + shape
            wfs = state.ibzwfs.wfs_qs[0][0]
            assert isinstance(wfs, PWFDWaveFunctions)
            dtype = wfs.psit_nX.data.dtype
            self.work_arrays = np.empty(shape, dtype)

        dS = state.ibzwfs.wfs_qs[0][0].setups.overlap_correction
        dH = state.potential.dH
        Ht = partial(hamiltonian.apply, state.potential.vt_sR)
        ibzwfs = state.ibzwfs
        error = 0.0
        for wfs in ibzwfs:
            e = self.iterate1(wfs, Ht, dH, dS)
            error += wfs.weight * e
        return ibzwfs.kpt_comm.sum(error) * ibzwfs.spin_degeneracy

    def iterate1(self, wfs, Ht, dH, dS):
        H_NN = self.H_NN
        S_NN = self.S_NN
        M_nn = self.M_nn

        psit_nX = wfs.psit_nX
        psit2_nX = psit_nX.new(data=self.work_arrays[0])
        psit3_nX = psit_nX.new(data=self.work_arrays[1])

        B = psit_nX.dims[0]  # number of bands
        eig_N = np.empty(2 * B)

        wfs.subspace_diagonalize(Ht, dH,
                                 work_array=psit2_nX.data,
                                 Htpsit_nX=psit3_nX)
        residual_nX = psit3_nX  # will become (H-e*S)|psit> later

        P_ani = wfs.P_ani
        P2_ani = P_ani.new()
        P3_ani = P_ani.new()

        domain_comm = psit_nX.desc.comm
        band_comm = psit_nX.comm
        is_domain_band_master = domain_comm.rank == 0 and band_comm.rank == 0

        M0_nn = M_nn
        assert band_comm.size == 1

        if domain_comm.rank == 0:
            eig_N[:B] = wfs.eig_n

        def me(a, b, function=None):
            """Matrix elements"""
            return a.matrix_elements(b,
                                     domain_sum=False,
                                     out=M_nn,
                                     function=function,
                                     cc=True)

        Ht = partial(Ht, out=residual_nX, spin=wfs.spin)
        dH = partial(dH, spin=wfs.spin)

        calculate_residuals(residual_nX, dH, dS, wfs, P2_ani, P3_ani)

        def copy(C_nn: Array2D) -> None:
            domain_comm.sum(M_nn.data, 0)
            if domain_comm.rank == 0:
                M_nn.redist(M0_nn)
                if band_comm.rank == 0:
                    C_nn[:] = M0_nn.data

        for i in range(self.niter):
            if i == self.niter - 1:  # last iteration
                # Calulate error before we destroy residuals:
                weight_n = calculate_weights(self.converge_bands, wfs)
                if weight_n is None:
                    error = np.inf
                else:
                    error = weight_n @ residual_nX.norm2()
                    if wfs.ncomponents == 4:
                        error = error.sum()

            self.preconditioner(psit_nX, residual_nX, out=psit2_nX)

            # Calculate projections
            wfs.pt_aiX.integrate(psit2_nX, out=P2_ani)

            # <psi2 | H | psi2>
            me(psit2_nX, psit2_nX, function=Ht)
            dH(P2_ani, out_ani=P3_ani)
            P2_ani.matrix.multiply(P3_ani, opb='C', symmetric=True, beta=1,
                                   out=M_nn)
            copy(H_NN.data[B:, B:])

            # <psi2 | H | psi>
            me(residual_nX, psit_nX)
            P3_ani.matrix.multiply(P_ani, opb='C', beta=1.0, out=M_nn)
            copy(H_NN.data[B:, :B])

            # <psi2 | S | psi2>
            me(psit2_nX, psit2_nX)
            dS(P2_ani, out_ani=P3_ani)
            P2_ani.matrix.multiply(P3_ani, opb='C', symmetric=True, beta=1,
                                   out=M_nn)
            copy(S_NN.data[B:, B:])

            # <psi2 | S | psi>
            me(psit2_nX, psit_nX)
            P3_ani.matrix.multiply(P_ani, opb='C', beta=1.0, out=M_nn)
            copy(S_NN.data[B:, :B])

            if is_domain_band_master:
                H_NN.data[:B, :B] = np.diag(eig_N[:B])
                S_NN.data[:B, :B] = np.eye(B)
                if debug:
                    H_NN.data[np.triu_indices(2 * B, 1)] = 42.0
                    S_NN.data[np.triu_indices(2 * B, 1)] = 42.0

                eig_N[:], H_NN.data[:] = eigh(H_NN.data, S_NN.data,
                                              lower=True,
                                              check_finite=debug,
                                              overwrite_b=True)
                wfs._eig_n = eig_N[:B]
            if domain_comm.rank == 0:
                band_comm.broadcast(wfs.eig_n, 0)
            domain_comm.broadcast(wfs.eig_n, 0)

            if domain_comm.rank == 0:
                if band_comm.rank == 0:
                    M0_nn.data[:] = H_NN.data[:B, :B]
                M0_nn.redist(M_nn)
            domain_comm.broadcast(M_nn.data, 0)

            M_nn.multiply(psit_nX, opa='C', out=residual_nX)
            M_nn.multiply(P_ani, opa='C', out=P3_ani)

            if domain_comm.rank == 0:
                if band_comm.rank == 0:
                    M0_nn.data[:] = H_NN.data[B:, :B]
                M0_nn.redist(M_nn)
            domain_comm.broadcast(M_nn.data, 0)

            M_nn.multiply(psit2_nX, opa='C', beta=1.0, out=residual_nX)
            M_nn.multiply(P2_ani, opa='C', beta=1.0, out=P3_ani)
            psit_nX.data[:] = residual_nX.data
            P_ani, P3_ani = P3_ani, P_ani
            wfs._P_ani = P_ani

            if i < self.niter - 1:
                Ht(psit_nX)
                calculate_residuals(residual_nX, dH, dS, wfs, P2_ani, P3_ani)

        return error


def calculate_residuals(residuals_nX: DA,
                        dH: AAFunc,
                        dS: AAFunc,
                        wfs: PWFDWaveFunctions,
                        P1_ani: AA,
                        P2_ani: AA) -> None:
    for r, e, p in zip(residuals_nX.data, wfs.myeig_n, wfs.psit_nX.data):
        axpy(-e, p, r)

    dH(wfs.P_ani, P1_ani)
    if wfs.ncomponents < 4:
        subscripts = 'nI, n -> nI'
    else:
        subscripts = 'nsI, n -> nsI'
    np.einsum(subscripts, wfs.P_ani.data, wfs.myeig_n, out=P2_ani.data)
    dS(P2_ani, P2_ani)
    P1_ani.data -= P2_ani.data
    wfs.pt_aiX.add_to(residuals_nX, P1_ani)


def calculate_weights(converge_bands: int | str,
                      wfs: PWFDWaveFunctions) -> Array1D | None:
    """Calculate convergence weights for all eigenstates."""
    if converge_bands == 'occupied':
        # Converge occupied bands:
        try:
            # Methfessel-Paxton distribution can give negative
            # occupation numbers - so we take the absolute value:
            return np.abs(wfs.occ_n)
        except ValueError:
            # No eigenvalues yet:
            return None

    if isinstance(converge_bands, int):
        # Converge fixed number of bands:
        n = converge_bands
        if wfs.psit_nX.comm.size > 1:
            raise NotImplementedError

        nbands = wfs.psit_nX.mydims[0]
        weight_n = np.zeros(nbands)
        if n < 0:
            n += nbands
        weight_n[:n] = 1.0
        return weight_n

    else:
        assert False
        # Converge state with energy up to CBM + delta:
        assert converge_bands.startswith('CBM+')
        # delta = float(converge_bands[4:]) / Ha
        return None

        """
        if wfs.kpt_u[0].f_n is None:
            weight_un[:] = np.inf  # no eigenvalues yet
        else:
            # Collect all eigenvalues and calculate band gap:
            efermi = np.mean(wfs.fermi_levels)
            eps_skn = np.array(
                [[wfs.collect_eigenvalues(k, spin) - efermi
                  for k in range(wfs.kd.nibzkpts)]
                 for spin in range(wfs.nspins)])
            if wfs.world.rank > 0:
                eps_skn = np.empty((wfs.nspins,
                                    wfs.kd.nibzkpts,
                                    wfs.bd.nbands))
            wfs.world.broadcast(eps_skn, 0)
            try:
                # Find bandgap + positions of CBM:
                gap, _, (s, k, n) = _bandgap(eps_skn,
                                             spin=None, direct=False)
            except ValueError:
                gap = 0.0

            if gap == 0.0:
                cbm = efermi
            else:
                cbm = efermi + eps_skn[s, k, n]

            ecut = cbm + delta

            for weight_n, kpt in zip(weight_un, wfs.kpt_u):
                weight_n[kpt.eps_n < ecut] = kpt.weight

            if (eps_skn[:, :, -1] < ecut - efermi).any():
                # We don't have enough bands!
                weight_un[:] = np.inf
        """
    return None


class EmptyMatrix:
    data = np.empty((0, 0))



# web-page: H2O_rraman_spectrum.png
import matplotlib.pyplot as plt
from ase import io
from ase.vibrations.placzek import Placzek
from gpaw.lrtddft import LrTDDFT

atoms = io.read('relaxed.traj')
pz = Placzek(atoms, LrTDDFT,
             name='ir',  # use ir-calculation for frequencies
             exname='rraman_erange17')  # use LrTDDFT for intensities

gamma = 0.2  # width

for i, omega in enumerate([2.41, 8]):  # photon energies
    plt.subplot(211 + i)
    x, y = pz.get_spectrum(omega, gamma,
                           start=1000, end=4000, type='Lorentzian')
    plt.text(0.1, 0.8, f'{omega} eV', transform=plt.gca().transAxes)
    plt.plot(x, y)
    plt.ylabel('cross section')

plt.xlabel('frequency [cm$^{-1}$]')
plt.savefig('H2O_rraman_spectrum.png')



# web-page: H2O_rraman_summary.txt
from ase import io
from ase.vibrations.placzek import Placzek
from gpaw.lrtddft import LrTDDFT

atoms = io.read('relaxed.traj')
pz = Placzek(atoms, LrTDDFT,
             name='ir',  # use ir-calculation for frequencies
             exname='rraman_erange17')  # use LrTDDFT for intensities

omega = 0  # excitation frequency
gamma = 0.2  # width
with open('H2O_rraman_summary.txt', 'w') as fd:
    pz.summary(omega, gamma, log=fd)



import numpy as np
from ase.units import Hartree, Bohr
from gpaw.external import ConstantElectricField
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.laser import GaussianPulse
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter

# Temporal shape of the time-dependent potential
pulse = GaussianPulse(1e-5, 10e3, 1.12, 0.3, 'sin')
# Spatial shape of the time-dependent potential
ext = ConstantElectricField(Hartree / Bohr, [1., 0., 0.])
# Full time-dependent potential
td_potential = {'ext': ext, 'laser': pulse}

# Write the temporal shape to a file
pulse.write('pulse.dat', np.arange(0, 30e3, 10.0))

# Set up the time-propagation calculation
td_calc = LCAOTDDFT('gs.gpw', td_potential=td_potential,
                    txt='tdpulse.out')

# Attach the data recording and analysis tools
DipoleMomentWriter(td_calc, 'dmpulse.dat')

# Propagate
td_calc.propagate(20, 1500)

# Save the state for restarting later
td_calc.write('tdpulse.gpw', mode='all')



from myqueue.workflow import run


def workflow():
    with run(script='gs.py', cores=24):
        td_jobs = []
        for kick in 'xyz':
            td = run(script='td.py', args=['--kick', kick],
                     cores=24, tmax='6h')
            td_jobs.append(td)
        with run(script='spec.py', deps=td_jobs):
            run(script='plot_spectrum.py')



import pytest
from ase import Atoms

from gpaw import GPAW, FermiDirac
from gpaw.mpi import serial_comm
from gpaw.xc.fxc import FXCCorrelation
from gpaw.xc.rpa import RPACorrelation


@pytest.mark.response
def test_ralda_ralda_energy_He(in_tmp_dir, scalapack):
    a = 3.0
    atoms = Atoms('He', cell=[a, a, a], pbc=True)
    calc = GPAW(mode=dict(name='pw', ecut=200),
                kpts=dict(size=(2, 2, 2), gamma=True),
                nbands=2,
                occupations=FermiDirac(0.001),
                # FXCCorrelation needs a serial-comm GPAW object:
                communicator=serial_comm)
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=20)

    ecuts = [20, 30]
    rpa = RPACorrelation(calc, nfrequencies=8)
    E_rpa1 = rpa.calculate(ecut=ecuts)[-1]

    def fxc(xc, nfrequencies=8, **kwargs):
        return FXCCorrelation(calc, xc=xc, **kwargs).calculate(ecut=ecuts)[-1]

    energies = [
        fxc('RPA', nlambda=16),
        fxc('rALDA', unit_cells=[1, 1, 2]),
        fxc('rAPBE', unit_cells=[1, 1, 2]),
        fxc('rALDA', av_scheme='wavevector'),
        fxc('rAPBE', av_scheme='wavevector'),
        fxc('JGMs', av_scheme='wavevector', Eg=3.1, nlambda=2),
        fxc('CP_dyn', av_scheme='wavevector', nfrequencies=2, nlambda=2)]

    assert E_rpa1 == pytest.approx(energies[0], abs=0.01)

    refs = [-0.1054,
            -0.0560,
            -0.0523,
            -0.0241,
            -0.0288,
            -0.0263,
            -0.0275]

    for val, ref in zip(energies, refs):
        assert val == pytest.approx(ref, abs=0.001)



import numpy as np
from ase import Atoms
import ase.units as units
from gpaw import GPAW, PW
from gpaw.hyperfine import hyperfine_parameters

h = Atoms('H', magmoms=[1])
h.center(vacuum=3)
h.calc = GPAW(mode=PW(400), txt=None)
e = h.get_potential_energy()
A = hyperfine_parameters(h.calc)[0] * 5.586
a = np.trace(A) / 3
frequency = a * units._e / units._hplanck  # Hz
wavelength = units._c / frequency  # meters
print(f'{wavelength * 100:.1f} cm')
assert abs(wavelength - 0.232) < 0.0005



import os
from pathlib import Path


def send_email(tasks):
    import smtplib
    from email.message import EmailMessage

    txt = 'Hi!\n\n'
    for task in tasks:
        if task.state in {'FAILED', 'CANCELED', 'TIMEOUT', 'MEMORY'}:
            id, dir, name, res, age, status, t, err = task.words()
            txt += ('test: {}/{}@{}: {}\ntime: {}\nerror: {}\n\n'
                    .format(dir.split('agts/gpaw')[1],
                            name,
                            res[:-1],
                            status,
                            t,
                            err))
    txt += 'Best regards,\nNiflheim\n'

    msg = EmailMessage()
    msg.set_content(txt)
    msg['Subject'] = 'Failing Niflheim-tests!'
    msg['From'] = 'agts@niflheim.dtu.dk'
    msg['To'] = 'jjmo@dtu.dk'
    s = smtplib.SMTP('smtp.ait.dtu.dk')
    s.send_message(msg)
    s.quit()


def find_created_files():
    names = set()
    for path in Path().glob('**/*.py'):
        if path.parts[0] == 'build':
            continue
        filenames = []
        for line in path.read_text().splitlines():
            if not line.startswith('# web-page:'):
                break
            filenames += line.split(':')[1].split(',')
        for name in filenames:
            name = name.strip()
            if name in names:
                raise RuntimeError(
                    f'The name {name!r} is used in more than one place!')
            names.add(name)
            yield path.with_name(name)


def collect_files_for_web_page():
    os.chdir('gpaw/doc')
    folder = Path('agts-files')
    if not folder.is_dir():
        folder.mkdir()
    for path in find_created_files():
        print(path)
        (folder / path.name).write_bytes(path.read_bytes())
    # os.environ['WEB_PAGE_FOLDER']



# web-page: Raman_all.png
import numpy as np
from gpaw import GPAW
from gpaw.raman.raman import (calculate_raman, calculate_raman_intensity,
                              plot_raman)

# Load pre-computed calculation
calc = GPAW('gs.gpw', parallel={'domain': 1, 'band': 1})
atoms = calc.atoms

# laser frequency 633 nm approx 1.958676 eV
w_l = 1.958676
suffix = '632nm'

# use previously saved phonon frequencies
w_ph = np.load('vib_frequencies.npy')

# Scan through all polarisations
pollist = []
for d_i in (0, 1, 2):
    for d_o in (0, 1, 2):
        # Calculate mode resolved Raman tensor for given direction
        calculate_raman(calc, w_ph, w_l, d_i, d_o, resonant_only=True,
                        suffix=suffix)
        if calc.wfs.kd.comm.rank == 0:
            # Calculate Raman intensity
            calculate_raman_intensity(w_ph, d_i, d_o, suffix=suffix)
            pollist.append('{}{}_{}'.format('xyz'[d_i], 'xyz'[d_o], suffix))

# And plot
if calc.wfs.kd.comm.rank == 0:
    plot_raman(figname='Raman_all.png', RIsuffix=pollist)



from gpaw.atom.generator import Generator
from gpaw.atom.basis import BasisMaker
from gpaw.atom.configurations import parameters, parameters_extra

atom = 'Ag'
xc = 'GLLBSC'
name = 'my'
if atom in parameters_extra:
    args = parameters_extra[atom]  # Choose the smaller setup
else:
    args = parameters[atom]  # Choose the larger setup
args.update(dict(name=name, exx=True))

# Generate setup
generator = Generator(atom, xc, scalarrel=True)
generator.run(write_xml=True, **args)

# Generate basis
bm = BasisMaker(atom, name=f'{name}.{xc}', xc=xc, run=False)
bm.generator.run(write_xml=False, **args)
basis = bm.generate(zetacount=2, polarizationcount=0,
                    jvalues=[0, 1, 2])  # include d, s and p
basis.write_xml()



import copy
from ase.build import molecule
from ase.parallel import paropen
from gpaw import GPAW
from gpaw.mom import prepare_mom_calculation

atoms = molecule('H2O')
atoms.center(vacuum=7)

calc = GPAW(mode='fd',
            basis='dzp',
            nbands=6,
            h=0.2,
            xc='PBE',
            spinpol=True,
            symmetry='off',
            convergence={'bands': -1},
            txt='h2o.txt')
atoms.calc = calc

# Ground-state calculation
E_gs = atoms.get_potential_energy()

# Ground-state occupation numbers
f_gs = []
for s in range(2):
    f_gs.append(calc.get_occupation_numbers(spin=s))

# Triplet n->3s occupation numbers
f_t = copy.deepcopy(f_gs)
f_t[0][2] -= 1.  # Remove one electron from homo-1 (n) spin up
f_t[1][4] += 1.  # Add one electron to lumo (3s) spin down

# MOM calculation for triplet n->3s state
prepare_mom_calculation(calc, atoms, f_t)
E_t = atoms.get_potential_energy()

# Mixed-spin n->3s occupation numbers
f_m = copy.deepcopy(f_gs)
f_m[0][2] -= 1.  # Remove one electron from homo-1 (n) spin up
f_m[0][4] += 1.  # Add one electron to lumo (3s) spin up

# MOM calculation for mixed-spin n->3s state
prepare_mom_calculation(calc, atoms, f_m)
E_m = atoms.get_potential_energy()
E_s = 2 * E_m - E_t  # Spin purified energy

with paropen('h2o_energies.txt', 'w') as fd:
    print(f'Excitation energy triplet n->3s state: {E_t - E_gs:.2f} eV',
          file=fd)
    print(f'Excitation energy singlet n->3s state: {E_s - E_gs:.2f} eV',
          file=fd)
    # https://doi.org/10.1021/acs.jctc.8b00406
    print('Experimental excitation energy triplet n->3s state: 9.46 eV',
          file=fd)
    print('Experimental excitation energy singlet n->3s state: 9.67 eV',
          file=fd)



from ase.vibrations.resonant_raman import ResonantRamanCalculator
from gpaw import GPAW, FermiDirac
from gpaw.cluster import Cluster
from gpaw.lrtddft import LrTDDFT

h = 0.25
xc = 'PBE'
atoms = Cluster('relaxed.traj')
atoms.minimal_box(4., h=h)

atoms.calc = GPAW(xc=xc, h=h, nbands=50,
                  occupations=FermiDirac(width=0.1),
                  eigensolver='cg', symmetry={'point_group': False},
                  convergence={'eigenstates': 1.e-5, 'bands': -10})
atoms.get_potential_energy()

erange = 17
ext = '_erange{0}'.format(erange)
gsname = exname = 'rraman' + ext
rr = ResonantRamanCalculator(
    atoms, LrTDDFT, name=gsname, exname=exname,
    exkwargs={'restrict': {'energy_range': erange, 'eps': 0.4}},)
rr.run()



import functools
import os
import pickle
import warnings
from math import pi
from pathlib import Path

import gpaw.mpi as mpi
import numpy as np
from ase.dft.kpoints import monkhorst_pack
from ase.parallel import paropen
from ase.units import Ha
from ase.utils import opencew, pickleload
from ase.utils.timing import timer
from gpaw import GPAW, debug
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.response.chi0 import Chi0Calculator
from gpaw.pw.descriptor import (PWDescriptor, PWMapping,
                                count_reciprocal_vectors)
from gpaw.response.fxckernel_calc import calculate_kernel
from gpaw.response.gamma_int import GammaIntegrator
from gpaw.response.hilbert import GWHilbertTransforms
from gpaw.response.kernels import get_coulomb_kernel, get_integrated_kernel
from gpaw.response.pair import NoCalculatorPairDensity
from gpaw.response.pw_parallelization import Blocks1D
from gpaw.response.q0_correction import Q0Correction
from gpaw.response.temp import DielectricFunctionCalculator
from gpaw.response.wstc import WignerSeitzTruncatedCoulomb
from gpaw.utilities.progressbar import ProgressBar
from gpaw.xc.exx import EXX, select_kpts
from gpaw.xc.fxc import XCFlags
from gpaw.xc.tools import vxc
from gpaw.response.context import calc_and_context


class Sigma:
    def __init__(self, esknshape):
        self._buf = np.zeros((2, * esknshape))
        # self-energies and derivatives:
        self.sigma_eskn, self.dsigma_eskn = self._buf

    def sum(self, comm):
        comm.sum(self._buf)


class G0W0Outputs:
    def __init__(self, fd, shape, ecut_e, sigma_eskn, dsigma_eskn,
                 eps_skn, vxc_skn, exx_skn, f_skn):
        self.extrapolate(fd, shape, ecut_e, sigma_eskn, dsigma_eskn)
        self.Z_skn = 1 / (1 - self.dsigma_skn)

        # G0W0 single-step.
        # If we want GW0 again, we need to grab the expressions
        # from e.g. e73917fca5b9dc06c899f00b26a7c46e7d6fa749
        # or earlier and use qp correctly.
        self.qp_skn = eps_skn + self.Z_skn * (
            -vxc_skn + exx_skn + self.sigma_skn)

        self.sigma_eskn = sigma_eskn
        self.dsigma_eskn = dsigma_eskn

        self.eps_skn = eps_skn
        self.vxc_skn = vxc_skn
        self.exx_skn = exx_skn
        self.f_skn = f_skn

    def extrapolate(self, fd, shape, ecut_e, sigma_eskn, dsigma_eskn):
        if len(ecut_e) == 1:
            self.sigma_skn = sigma_eskn[0]
            self.dsigma_skn = dsigma_eskn[0]
            self.sigr2_skn = None
            self.dsigr2_skn = None
            return

        from scipy.stats import linregress

        # Do linear fit of selfenergy vs. inverse of number of plane waves
        # to extrapolate to infinite number of plane waves

        print('', file=fd)
        print('Extrapolating selfenergy to infinite energy cutoff:',
              file=fd)
        print('  Performing linear fit to %d points' % len(ecut_e),
              file=fd)
        self.sigr2_skn = np.zeros(shape)
        self.dsigr2_skn = np.zeros(shape)
        self.sigma_skn = np.zeros(shape)
        self.dsigma_skn = np.zeros(shape)
        invN_i = ecut_e**(-3. / 2)
        for m in range(np.product(shape)):
            s, k, n = np.unravel_index(m, shape)

            slope, intercept, r_value, p_value, std_err = \
                linregress(invN_i, sigma_eskn[:, s, k, n])

            self.sigr2_skn[s, k, n] = r_value**2
            self.sigma_skn[s, k, n] = intercept

            slope, intercept, r_value, p_value, std_err = \
                linregress(invN_i, dsigma_eskn[:, s, k, n])

            self.dsigr2_skn[s, k, n] = r_value**2
            self.dsigma_skn[s, k, n] = intercept

        if np.any(self.sigr2_skn < 0.9) or np.any(self.dsigr2_skn < 0.9):
            print('  Warning: Bad quality of linear fit for some (n,k). ',
                  file=fd)
            print('           Higher cutoff might be necesarry.', file=fd)

        print('  Minimum R^2 = %1.4f. (R^2 Should be close to 1)' %
              min(np.min(self.sigr2_skn), np.min(self.dsigr2_skn)),
              file=fd)

    def get_results_eV(self):
        results = {
            'f': self.f_skn,
            'eps': self.eps_skn * Ha,
            'vxc': self.vxc_skn * Ha,
            'exx': self.exx_skn * Ha,
            'sigma': self.sigma_skn * Ha,
            'dsigma': self.dsigma_skn,
            'Z': self.Z_skn,
            'qp': self.qp_skn * Ha}

        results.update(
            sigma_eskn=self.sigma_eskn * Ha,
            dsigma_eskn=self.dsigma_eskn)

        if self.sigr2_skn is not None:
            assert self.dsigr2_skn is not None
            results['sigr2_skn'] = self.sigr2_skn
            results['dsigr2_skn'] = self.dsigr2_skn

        return results


class QSymmetryOp:
    def __init__(self, symno, U_cc, sign):
        self.symno = symno
        self.U_cc = U_cc
        self.sign = sign

    def apply(self, q_c):
        return self.sign * (self.U_cc @ q_c)

    def check_q_Q_symmetry(self, Q_c, q_c):
        d_c = self.apply(q_c) - Q_c
        assert np.allclose(d_c.round(), d_c)

    def get_shift0(self, q_c, Q_c):
        shift0_c = q_c - self.apply(Q_c)
        assert np.allclose(shift0_c.round(), shift0_c)
        return shift0_c.round().astype(int)

    def get_M_vv(self, cell_cv):
        # We'll be inverting these cells a lot.
        # Should have an object with the cell and its inverse which does this.
        return cell_cv.T @ self.U_cc.T @ np.linalg.inv(cell_cv).T

    @classmethod
    def get_symops(cls, qd, iq, q_c):
        # Loop over all k-points in the BZ and find those that are
        # related to the current IBZ k-point by symmetry
        Q1 = qd.ibz2bz_k[iq]
        done = set()
        for Q2 in qd.bz2bz_ks[Q1]:
            if Q2 >= 0 and Q2 not in done:
                time_reversal = qd.time_reversal_k[Q2]
                symno = qd.sym_k[Q2]
                Q_c = qd.bzk_kc[Q2]

                symop = cls(
                    symno=symno,
                    U_cc=qd.symmetry.op_scc[symno],
                    sign=1 - 2 * time_reversal)

                symop.check_q_Q_symmetry(Q_c, q_c)
                # Q_c, symop = QSymmetryOp.from_qd(qd, Q2, q_c)
                yield Q_c, symop
                done.add(Q2)


gw_logo = """\
  ___  _ _ _
 |   || | | |
 | | || | | |
 |__ ||_____|
 |___|
"""


def get_max_nblocks(world, calc, ecut):
    nblocks = world.size
    if not isinstance(calc, (str, Path)):
        raise Exception('Using a calulator is not implemented at '
                        'the moment, load from file!')
        # nblocks_calc = calc
    else:
        nblocks_calc = GPAW(calc)
    ngmax = []
    for q_c in nblocks_calc.wfs.kd.bzk_kc:
        qd = KPointDescriptor([q_c])
        pd = PWDescriptor(np.min(ecut) / Ha,
                          nblocks_calc.wfs.gd, complex, qd)
        ngmax.append(pd.ngmax)
    nG = np.min(ngmax)

    while nblocks > nG**0.5 + 1 or world.size % nblocks != 0:
        nblocks -= 1

    mynG = (nG + nblocks - 1) // nblocks
    assert mynG * (nblocks - 1) < nG
    return nblocks


def get_frequencies(frequencies, domega0, omega2):
    if domega0 is not None or omega2 is not None:
        assert frequencies is None
        frequencies = {'type': 'nonlinear',
                       'domega0': 0.025 if domega0 is None else domega0,
                       'omega2': 10.0 if omega2 is None else omega2}
        warnings.warn(f'Please use frequencies={frequencies}')
    elif frequencies is None:
        frequencies = {'type': 'nonlinear',
                       'domega0': 0.025,
                       'omega2': 10.0}
    else:
        assert frequencies['type'] == 'nonlinear'
    return frequencies


def get_qdescriptor(kd, atoms):
    # Find q-vectors and weights in the IBZ:
    assert -1 not in kd.bz2bz_ks
    offset_c = 0.5 * ((kd.N_c + 1) % 2) / kd.N_c
    bzq_qc = monkhorst_pack(kd.N_c) + offset_c
    qd = KPointDescriptor(bzq_qc)
    qd.set_symmetry(atoms, kd.symmetry)
    return qd


def choose_ecut_things(ecut, ecut_extrapolation):
    if ecut_extrapolation is True:
        pct = 0.8
        necuts = 3
        ecut_e = ecut * (1 + (1. / pct - 1) * np.arange(necuts)[::-1] /
                         (necuts - 1))**(-2 / 3)
    elif isinstance(ecut_extrapolation, (list, np.ndarray)):
        ecut_e = np.array(np.sort(ecut_extrapolation))
        ecut = ecut_e[-1]
    else:
        ecut_e = np.array([ecut])
    return ecut, ecut_e


class G0W0Calculator:
    def __init__(self, filename='gw', *,
                 chi0calc,
                 restartfile=None,
                 kpts, bands, nbands=None, ppa,
                 xckernel,
                 fxc_mode='GW', do_GW_too=False,
                 truncation=None, integrate_gamma=0,
                 eta, E0,
                 ecut_e,
                 frequencies=None,
                 q0_correction=False,
                 savepckl=True,
                 context):

        """G0W0 calculator.

        The G0W0 calculator is used is used to calculate the quasi
        particle energies through the G0W0 approximation for a number
        of states.

        Parameters
        ----------
        calc:
            GPAW calculator object or filename of saved calculator object.
        filename: str
            Base filename of output files.
        restartfile: str
            File that stores data necessary to restart a calculation.
        kpts: list
            List of indices of the IBZ k-points to calculate the quasi particle
            energies for.
        bands:
            Range of band indices, like (n1, n2), to calculate the quasi
            particle energies for. Bands n where n1<=n<n2 will be
            calculated.  Note that the second band index is not included.
        relbands:
            Same as *bands* except that the numbers are relative to the
            number of occupied bands.
            E.g. (-1, 1) will use HOMO+LUMO.
        frequencies:
            Input parameters for frequency_grid.
            Can be array of frequencies to evaluate the response function at
            or dictionary of parameters for build-in nonlinear grid
            (see :ref:`frequency grid`).
        ecut: float
            Plane wave cut-off energy in eV.
        ecut_extrapolation: bool or list
            If set to True an automatic extrapolation of the selfenergy to
            infinite cutoff will be performed based on three points
            for the cutoff energy.
            If an array is given, the extrapolation will be performed based on
            the cutoff energies given in the array.
        nbands: int
            Number of bands to use in the calculation. If None, the number will
            be determined from :ecut: to yield a number close to the number of
            plane waves used.
        ppa: bool
            Sets whether the Godby-Needs plasmon-pole approximation for the
            dielectric function should be used.
        xc: str
            Kernel to use when including vertex corrections.
        fxc_mode: str
            Where to include the vertex corrections; polarizability and/or
            self-energy. 'GWP': Polarizability only, 'GWS': Self-energy only,
            'GWG': Both.
        do_GW_too: bool
            When carrying out a calculation including vertex corrections, it
            is possible to get the standard GW results at the same time
            (almost for free).
        Eg: float
            Gap to apply in the 'JGMs' (simplified jellium-with-gap) kernel.
            If None the DFT gap is used.
        truncation: str
            Coulomb truncation scheme. Can be either wigner-seitz,
            2D, 1D, or 0D
        integrate_gamma: int
            Method to integrate the Coulomb interaction. 1 is a numerical
            integration at all q-points with G=[0,0,0] - this breaks the
            symmetry slightly. 0 is analytical integration at q=[0,0,0] only -
            this conserves the symmetry. integrate_gamma=2 is the same as 1,
            but the average is only carried out in the non-periodic directions.
        E0: float
            Energy (in eV) used for fitting in the plasmon-pole approximation.
        q0_correction: bool
            Analytic correction to the q=0 contribution applicable to 2D
            systems.
        nblocks: int
            Number of blocks chi0 should be distributed in so each core
            does not have to store the entire matrix. This is to reduce
            memory requirement. nblocks must be less than or equal to the
            number of processors.
        nblocksmax: bool
            Cuts chi0 into as many blocks as possible to reduce memory
            requirements as much as possible.
        savepckl: bool
            Save output to a pckl file.
        """

        # Note: self.wd should be our only representation of the frequencies.
        # We should therefore get rid of self.frequencies.
        # It is currently only used by the restart code,
        # so should be easy to remove after some further adaptation.
        self.frequencies = frequencies

        self.ecut_e = ecut_e / Ha

        self.context = context
        self.chi0calc = chi0calc
        self.pair = chi0calc.pair
        self.wd = chi0calc.wd
        self.gs = chi0calc.gs

        if ppa and self.pair.nblocks > 1:
            raise ValueError(
                'PPA is currently not compatible with block parallelisation.')

        self.timer = self.context.timer
        self.fd = self.context.fd
        self.blockcomm = self.pair.blockcomm
        self.world = self.context.world

        print(gw_logo, file=self.fd)

        self.xckernel = xckernel
        self.fxc_mode = fxc_mode
        self.do_GW_too = do_GW_too

        if not self.fxc_mode == 'GW':
            assert self.xckernel.xc != 'RPA'

        if self.do_GW_too:
            assert self.xckernel.xc != 'RPA'
            assert self.fxc_mode != 'GW'
            if restartfile is not None:
                raise RuntimeError('Restart function does not currently work '
                                   'with do_GW_too=True.')

        self.filename = filename
        self.restartfile = restartfile
        self.savepckl = savepckl
        self.ppa = ppa
        self.truncation = truncation
        self.integrate_gamma = integrate_gamma
        self.eta = eta / Ha
        self.E0 = E0 / Ha

        self.kpts = kpts
        self.bands = bands

        b1, b2 = self.bands
        self.shape = (self.gs.nspins, len(self.kpts), b2 - b1)

        self.nbands = nbands

        if self.gs.nspins != 1 and self.fxc_mode != 'GW':
            raise RuntimeError('Including a xc kernel does currently not '
                               'work for spinpolarized systems.')

        self.pair_distribution = self.pair.distribute_k_points_and_bands(
            b1, b2, self.gs.kd.ibz2bz_k[self.kpts])

        self.qd = get_qdescriptor(self.gs.kd, self.gs.atoms)

        if q0_correction:
            assert self.truncation == '2D'
            self.q0_corrector = Q0Correction(
                cell_cv=self.gs.gd.cell_cv, bzk_kc=self.gs.kd.bzk_kc,
                N_c=self.qd.N_c)
        else:
            self.q0_corrector = None

        self.print_parameters(kpts, b1, b2)
        self.fd.flush()
        self.hilbert_transform = None  # initialized when we create Chi0

        if self.ppa:
            print('Using Godby-Needs plasmon-pole approximation:',
                  file=self.fd)
            print('  Fitting energy: i*E0, E0 = %.3f Hartee' % self.E0,
                  file=self.fd)
        else:
            print('Using full frequency integration', file=self.fd)

    def print_parameters(self, kpts, b1, b2):
        p = functools.partial(print, file=self.fd)
        p()
        p('Quasi particle states:')
        if kpts is None:
            p('All k-points in IBZ')
        else:
            kptstxt = ', '.join(['{0:d}'.format(k) for k in self.kpts])
            p('k-points (IBZ indices): [' + kptstxt + ']')
        p('Band range: ({0:d}, {1:d})'.format(b1, b2))
        p()
        p('Computational parameters:')
        if len(self.ecut_e) == 1:
            p('Plane wave cut-off: {0:g} eV'.format(self.chi0calc.ecut * Ha))
        else:
            assert len(self.ecut_e) > 1
            p('Extrapolating to infinite plane wave cut-off using points at:')
            p('  [%.3f, %.3f, %.3f] eV' % tuple(self.ecut_e * Ha))
        p('Number of bands: {0:d}'.format(self.nbands))
        p('Coulomb cutoff:', self.truncation)
        p('Broadening: {0:g} eV'.format(self.eta * Ha))
        p()
        p('fxc mode:', self.fxc_mode)
        p('Kernel:', self.xckernel.xc)
        p('Do GW too:', self.do_GW_too)
        p()

    def get_eps_and_occs(self):
        eps_skn = np.empty(self.shape)  # KS-eigenvalues
        f_skn = np.empty(self.shape)  # occupation numbers

        nspins = self.gs.nspins
        b1, b2 = self.bands
        for i, k in enumerate(self.kpts):
            for s in range(nspins):
                u = s + k * nspins
                kpt = self.gs.kpt_u[u]
                eps_skn[s, i] = kpt.eps_n[b1:b2]
                f_skn[s, i] = kpt.f_n[b1:b2] / kpt.weight

        return eps_skn, f_skn

    @timer('G0W0')
    def calculate(self):
        """Starts the G0W0 calculation.

        Returns a dict with the results with the following key/value pairs:

        ===========  =============================================
        key          value
        ===========  =============================================
        ``f``        Occupation numbers
        ``eps``      Kohn-Sham eigenvalues in eV
        ``vxc``      Exchange-correlation
                     contributions in eV
        ``exx``      Exact exchange contributions in eV
        ``sigma``    Self-energy contributions in eV
        ``dsigma``   Self-energy derivatives
        ``sigma_e``  Self-energy contributions in eV
                     used for ecut extrapolation
        ``Z``        Renormalization factors
        ``qp``       Quasi particle (QP) energies in eV
        ``iqp``      GW0/GW: QP energies for each iteration in eV
        ===========  =============================================

        All the values are ``ndarray``'s of shape
        (spins, IBZ k-points, bands)."""

        loaded = False
        if self.restartfile is not None:
            loaded = self.load_restart_file()
            if not loaded:
                self.last_q = -1
                self.previous_sigma = 0.
                self.previous_dsigma = 0.

            else:
                print('Reading ' + str(self.last_q + 1) +
                      ' q-point(s) from the previous calculation: ' +
                      self.restartfile + '.sigma.pckl', file=self.fd)
        else:
            self.last_q = -1
            self.previous_sigma = 0.
            self.previous_dsigma = 0.

        self.fd.flush()

        # Reset calculation
        sigmashape = (len(self.ecut_e), *self.shape)

        self.sigmas = {fxc_mode: Sigma(sigmashape)
                       for fxc_mode in self.fxc_modes}
        # Loop over q in the IBZ:
        print('Summing all q:', file=self.fd)
        pb = ProgressBar(self.fd)
        for nQ, (ie, pd0, Wdict, q_c, m2, symop, blocks1d, Q_aGii) in \
                enumerate(self.calculate_screened_potential()):

            for progress, kpt1, kpt2 in self.pair_distribution.kpt_pairs_by_q(
                    q_c, 0, m2):
                pb.update((nQ + progress) / self.qd.mynk)

                k1 = self.gs.kd.bz2ibz_k[kpt1.K]
                i = self.kpts.index(k1)

                self.calculate_q(ie, i, kpt1, kpt2, pd0, Wdict,
                                 symop=symop,
                                 sigmas=self.sigmas,
                                 blocks1d=blocks1d,
                                 Q_aGii=Q_aGii)
        pb.finish()

        return self.postprocess(self.sigmas, loaded)

    def postprocess(self, sigmas, loaded):
        all_results = {}
        for fxc_mode, sigma in sigmas.items():
            all_results[fxc_mode] = self.postprocess_single(fxc_mode, sigma,
                                                            loaded)
        self.all_results = all_results
        self.print_results(self.all_results)

        # After we have written the results restartfile is obsolete
        if self.restartfile is not None:
            if self.world.rank == 0:
                if os.path.isfile(self.restartfile + '.sigma.pckl'):
                    os.remove(self.restartfile + '.sigma.pckl')

        return self.results  # XXX ugly discrepancy

    def postprocess_single(self, fxc_name, sigma, loaded):
        sigma.sum(self.world)  # (Not so pretty that we finalize the sum here)

        if self.restartfile is not None and loaded:
            assert not self.do_GW_too
            sigma.sigma_eskn += self.previous_sigma
            sigma.dsigma_eskn += self.previous_dsigma

        output = self.calculate_g0w0_outputs(sigma)
        result = output.get_results_eV()

        if self.savepckl:
            with paropen(f'{self.filename}_results_{fxc_name}.pckl',
                         'wb') as fd:
                pickle.dump(result, fd, 2)

        return result

    @property
    def results_GW(self):
        if self.do_GW_too:
            return self.all_results['GW']

    @property
    def results(self):
        return self.all_results[self.fxc_mode]

    def calculate_q(self, ie, k, kpt1, kpt2, pd0, Wdict,
                    *, symop, sigmas, blocks1d, Q_aGii):
        """Calculates the contribution to the self-energy and its derivative
        for a given set of k-points, kpt1 and kpt2."""

        N_c = pd0.gd.N_c
        i_cG = symop.apply(np.unravel_index(pd0.Q_qG[0], N_c))

        q_c = self.gs.kd.bzk_kc[kpt2.K] - self.gs.kd.bzk_kc[kpt1.K]

        shift0_c = symop.get_shift0(q_c, pd0.kd.bzk_kc[0])
        shift_c = kpt1.shift_c - kpt2.shift_c - shift0_c

        I_G = np.ravel_multi_index(i_cG + shift_c[:, None], N_c, 'wrap')

        G_Gv = pd0.get_reciprocal_vectors()

        pos_av = np.dot(self.pair.spos_ac, pd0.gd.cell_cv)
        M_vv = symop.get_M_vv(pd0.gd.cell_cv)

        myQ_aGii = []
        for a, Q_Gii in enumerate(Q_aGii):
            x_G = np.exp(1j * np.dot(G_Gv, (pos_av[a] -
                                            np.dot(M_vv, pos_av[a]))))
            U_ii = self.gs.setups[a].R_sii[symop.symno]
            Q_Gii = np.dot(np.dot(U_ii, Q_Gii * x_G[:, None, None]),
                           U_ii.T).transpose(1, 0, 2)
            if symop.sign == -1:
                Q_Gii = Q_Gii.conj()
            myQ_aGii.append(Q_Gii)

        if debug:
            self.check(ie, i_cG, shift0_c, N_c, q_c, myQ_aGii)

        if self.ppa:
            calculate_sigma = self.calculate_sigma_ppa
        else:
            calculate_sigma = self.calculate_sigma

        for n in range(kpt1.n2 - kpt1.n1):
            ut1cc_R = kpt1.ut_nR[n].conj()
            eps1 = kpt1.eps_n[n]
            C1_aGi = [np.dot(Qa_Gii, P1_ni[n].conj())
                      for Qa_Gii, P1_ni in zip(myQ_aGii, kpt1.P_ani)]
            n_mG = self.pair.calculate_pair_densities(
                ut1cc_R, C1_aGi, kpt2, pd0, I_G)
            if symop.sign == 1:
                n_mG = n_mG.conj()

            f_m = kpt2.f_n
            deps_m = eps1 - kpt2.eps_n

            nn = kpt1.n1 + n - self.bands[0]

            assert set(Wdict) == set(sigmas)
            for fxc_mode in self.fxc_modes:
                sigma = sigmas[fxc_mode]
                W = Wdict[fxc_mode]
                sigma_contrib, dsigma_contrib = calculate_sigma(
                    n_mG, deps_m, f_m, W, blocks1d)
                sigma.sigma_eskn[ie, kpt1.s, k, nn] += sigma_contrib
                sigma.dsigma_eskn[ie, kpt1.s, k, nn] += dsigma_contrib

    def check(self, ie, i_cG, shift0_c, N_c, q_c, Q_aGii):
        I0_G = np.ravel_multi_index(i_cG - shift0_c[:, None], N_c, 'wrap')
        qd1 = KPointDescriptor([q_c])
        pd1 = PWDescriptor(self.ecut_e[ie], self.gs.gd, complex, qd1)
        G_I = np.empty(N_c.prod(), int)
        G_I[:] = -1
        I1_G = pd1.Q_qG[0]
        G_I[I1_G] = np.arange(len(I0_G))
        G_G = G_I[I0_G]
        assert len(I0_G) == len(I1_G)
        assert (G_G >= 0).all()
        for a, Q_Gii in enumerate(self.pair.initialize_paw_corrections(pd1)):
            e = abs(Q_aGii[a] - Q_Gii[G_G]).max()
            assert e < 1e-12

    @timer('Sigma')
    def calculate_sigma(self, n_mG, deps_m, f_m, C_swGG, blocks1d):
        """Calculates a contribution to the self-energy and its derivative for
        a given (k, k-q)-pair from its corresponding pair-density and
        energy."""
        o_m = abs(deps_m)
        # Add small number to avoid zeros for degenerate states:
        sgn_m = np.sign(deps_m + 1e-15)

        # Pick +i*eta or -i*eta:
        s_m = (1 + sgn_m * np.sign(0.5 - f_m)).astype(int) // 2

        w_m = self.wd.get_floor_index(o_m, safe=False)
        m_inb = np.where(w_m < len(self.wd) - 1)[0]
        o1_m = np.empty(len(o_m))
        o2_m = np.empty(len(o_m))
        o1_m[m_inb] = self.wd.omega_w[w_m[m_inb]]
        o2_m[m_inb] = self.wd.omega_w[w_m[m_inb] + 1]

        x = 1.0 / (self.qd.nbzkpts * 2 * pi * self.gs.volume)
        sigma = 0.0
        dsigma = 0.0
        # Performing frequency integration
        for o, o1, o2, sgn, s, w, n_G in zip(o_m, o1_m, o2_m,
                                             sgn_m, s_m, w_m, n_mG):

            if w >= len(self.wd.omega_w) - 1:
                continue

            C1_GG = C_swGG[s][w]
            C2_GG = C_swGG[s][w + 1]
            p = x * sgn
            myn_G = n_G[blocks1d.myslice]

            sigma1 = p * np.dot(np.dot(myn_G, C1_GG), n_G.conj()).imag
            sigma2 = p * np.dot(np.dot(myn_G, C2_GG), n_G.conj()).imag
            sigma += ((o - o1) * sigma2 + (o2 - o) * sigma1) / (o2 - o1)
            dsigma += sgn * (sigma2 - sigma1) / (o2 - o1)

        return sigma, dsigma

    def calculate_screened_potential(self):
        """Calculates the screened potential for each q-point in the 1st BZ.
        Since many q-points are related by symmetry, the actual calculation is
        only done for q-points in the IBZ and the rest are obtained by symmetry
        transformations. Results are returned as a generator to that it is not
        necessary to store a huge matrix for each q-point in the memory."""
        # The decorator $timer('W') doesn't work for generators, do we will
        # have to manually start and stop the timer here:
        self.timer.start('W')
        print('\nCalculating screened Coulomb potential', file=self.fd)
        if self.truncation is not None:
            print('Using %s truncated Coloumb potential' % self.truncation,
                  file=self.fd)

        chi0calc = self.chi0calc

        if self.truncation == 'wigner-seitz':
            wstc = WignerSeitzTruncatedCoulomb(
                self.gs.gd.cell_cv,
                self.gs.kd.N_c,
                chi0calc.fd)
        else:
            wstc = None

        self.hilbert_transform = GWHilbertTransforms(
            self.wd.omega_w, self.eta)
        print(self.wd, file=self.fd)

        # Find maximum size of chi-0 matrices:
        nGmax = max(count_reciprocal_vectors(chi0calc.ecut, self.gs.gd, q_c)
                    for q_c in self.qd.ibzk_kc)
        nw = len(self.wd)

        size = self.blockcomm.size

        mynGmax = (nGmax + size - 1) // size
        mynw = (nw + size - 1) // size

        # some memory sizes...
        if self.world.rank == 0:
            siz = (nw * mynGmax * nGmax +
                   max(mynw * nGmax, nw * mynGmax) * nGmax) * 16
            sizA = (nw * nGmax * nGmax + nw * nGmax * nGmax) * 16
            print('  memory estimate for chi0: local=%.2f MB, global=%.2f MB'
                  % (siz / 1024**2, sizA / 1024**2), file=self.fd)
            self.fd.flush()

        # Need to pause the timer in between iterations
        self.timer.stop('W')
        for iq, q_c in enumerate(self.qd.ibzk_kc):
            if iq <= self.last_q:
                continue

            if len(self.ecut_e) > 1:
                chi0bands = chi0calc.create_chi0(q_c, extend_head=False)
            else:
                chi0bands = None

            m1 = chi0calc.nocc1
            for ie, ecut in enumerate(self.ecut_e):
                self.timer.start('W')

                # First time calculation
                if ecut == chi0calc.ecut:
                    # Nothing to cut away:
                    m2 = self.nbands
                else:
                    m2 = int(self.gs.volume * ecut**1.5 * 2**0.5 / 3 / pi**2)
                    if m2 > self.nbands:
                        raise ValueError(f'Trying to extrapolate ecut to'
                                         f'larger number of bands ({m2})'
                                         f' than there are bands '
                                         f'({self.nbands}).')
                pdi, Wdict, blocks1d, Q_aGii = self.calculate_w(
                    chi0calc, q_c, chi0bands,
                    m1, m2, ecut, wstc, iq)
                m1 = m2

                self.timer.stop('W')

                for Q_c, symop in QSymmetryOp.get_symops(self.qd, iq, q_c):
                    yield (ie, pdi, Wdict, Q_c, m2, symop,
                           blocks1d, Q_aGii)

                if self.restartfile is not None:
                    self.save_restart_file(iq)

    @property
    def fxc_modes(self):
        modes = [self.fxc_mode]
        if self.do_GW_too:
            modes.append('GW')
        return modes

    @timer('WW')
    def calculate_w(self, chi0calc, q_c, chi0bands,
                    m1, m2, ecut, wstc,
                    iq):
        """Calculates the screened potential for a specified q-point."""

        chi0 = chi0calc.create_chi0(q_c, extend_head=False)
        chi0calc.fd = self.fd
        chi0calc.print_chi(chi0.pd)
        chi0calc.update_chi0(chi0, m1, m2, range(self.gs.nspins))

        if len(self.ecut_e) > 1:
            # Add chi from previous cutoff with remaining bands
            chi0.chi0_wGG += chi0bands.chi0_wGG
            chi0bands.chi0_wGG[:] = chi0.chi0_wGG.copy()
            if chi0.optical_limit:
                chi0.chi0_wxvG += chi0bands.chi0_wxvG
                chi0bands.chi0_wxvG[:] = chi0.chi0_wxvG.copy()
                chi0.chi0_wvv += chi0bands.chi0_wvv
                chi0bands.chi0_wvv[:] = chi0.chi0_wvv.copy()

        Wdict = {}

        for fxc_mode in self.fxc_modes:
            pdi, blocks1d, W_wGG = self.dyson_and_W_old(
                wstc, iq, q_c, chi0calc, chi0, ecut, Q_aGii=chi0calc.Q_aGii,
                fxc_mode=fxc_mode)

            if self.ppa:
                W_xwGG = W_wGG  # (ppa API is nonsense)
            else:
                with self.timer('Hilbert'):
                    W_xwGG = self.hilbert_transform(W_wGG)

            Wdict[fxc_mode] = W_xwGG

        return pdi, Wdict, blocks1d, chi0calc.Q_aGii

    def dyson_and_W_new(self, wstc, iq, q_c, chi0calc, chi0, ecut):
        assert not self.ppa
        assert not self.do_GW_too
        assert ecut == chi0.pd.ecut
        assert self.fxc_mode == 'GW'

        assert not np.allclose(q_c, 0)

        nW = len(self.wd)
        nG = chi0.pd.ngmax

        from gpaw.response.wgg import Grid

        WGG = (nW, nG, nG)
        WgG_grid = Grid(
            comm=self.blockcomm,
            shape=WGG,
            cpugrid=(1, self.blockcomm.size, 1))
        assert chi0.chi0_wGG.shape == WgG_grid.myshape

        my_gslice = WgG_grid.myslice[1]

        dielectric_WgG = chi0.chi0_wGG  # XXX
        for iw, chi0_GG in enumerate(chi0.chi0_wGG):
            sqrtV_G = get_coulomb_kernel(chi0.pd,  # XXX was: pdi
                                         self.gs.kd.N_c,
                                         truncation=self.truncation,
                                         wstc=wstc)**0.5
            e_GG = np.eye(nG) - chi0_GG * sqrtV_G * sqrtV_G[:, np.newaxis]
            e_gG = e_GG[my_gslice]

            dielectric_WgG[iw, :, :] = e_gG

        wgg_grid = Grid(comm=self.blockcomm, shape=WGG)

        dielectric_wgg = wgg_grid.zeros(dtype=complex)
        WgG_grid.redistribute(wgg_grid, dielectric_WgG, dielectric_wgg)

        assert np.allclose(dielectric_wgg, dielectric_WgG)

        wgg_grid.invert_inplace(dielectric_wgg)

        wgg_grid.redistribute(WgG_grid, dielectric_wgg, dielectric_WgG)
        inveps_WgG = dielectric_WgG

        self.timer.start('Dyson eq.')

        for iw, inveps_gG in enumerate(inveps_WgG):
            inveps_gG -= np.identity(nG)[my_gslice]
            thing_GG = sqrtV_G * sqrtV_G[:, np.newaxis]
            inveps_gG *= thing_GG[my_gslice]

        W_WgG = inveps_WgG
        Wp_wGG = W_WgG.copy()
        Wm_wGG = W_WgG.copy()
        return chi0.pd, Wm_wGG, Wp_wGG  # not Hilbert transformed yet

    def dyson_and_W_old(self, wstc, iq, q_c, chi0calc, chi0,
                        ecut, Q_aGii, fxc_mode):
        nG = chi0.pd.ngmax
        blocks1d = chi0.blocks1d

        wblocks1d = Blocks1D(self.blockcomm, len(self.wd))

        # The copy() is only required when doing GW_too, since we need
        # to run this whole thin twice.
        chi0_wGG = chi0.blockdist.redistribute(chi0.chi0_wGG.copy(), chi0.nw)

        pd = chi0.pd
        chi0_wxvG = chi0.chi0_wxvG
        chi0_wvv = chi0.chi0_wvv

        if ecut == pd.ecut:
            pdi = pd
            G2G = None

        elif ecut < pd.ecut:  # construct subset chi0 matrix with lower ecut
            pdi = PWDescriptor(ecut, pd.gd, dtype=pd.dtype,
                               kd=pd.kd)
            nG = pdi.ngmax
            blocks1d = Blocks1D(self.blockcomm, nG)
            G2G = PWMapping(pdi, pd).G2_G1
            chi0_wGG = chi0_wGG.take(G2G, axis=1).take(G2G, axis=2)

            if chi0_wxvG is not None:
                chi0_wxvG = chi0_wxvG.take(G2G, axis=3)

            if Q_aGii is not None:
                for a, Q_Gii in enumerate(Q_aGii):
                    Q_aGii[a] = Q_Gii.take(G2G, axis=0)

        if self.integrate_gamma != 0:
            reduced = (self.integrate_gamma == 2)
            V0, sqrtV0 = get_integrated_kernel(pdi,
                                               self.gs.kd.N_c,
                                               truncation=self.truncation,
                                               reduced=reduced,
                                               N=100)
        elif self.integrate_gamma == 0 and np.allclose(q_c, 0):
            bzvol = (2 * np.pi)**3 / self.gs.volume / self.qd.nbzkpts
            Rq0 = (3 * bzvol / (4 * np.pi))**(1. / 3.)
            V0 = 16 * np.pi**2 * Rq0 / bzvol
            sqrtV0 = (4 * np.pi)**(1.5) * Rq0**2 / bzvol / 2

        delta_GG = np.eye(nG)

        if self.ppa:
            einv_wGG = []

        if fxc_mode == 'GW':
            fv = delta_GG
        else:
            fv = self.xckernel.calculate(nG, iq, G2G)

        # Generate fine grid in vicinity of gamma
        kd = self.gs.kd
        if np.allclose(q_c, 0) and len(chi0_wGG) > 0:
            gamma_int = GammaIntegrator(truncation=self.truncation,
                                        kd=kd, pd=pd,
                                        chi0_wvv=chi0_wvv[wblocks1d.myslice],
                                        chi0_wxvG=chi0_wxvG[wblocks1d.myslice])

        self.timer.start('Dyson eq.')

        def get_sqrtV_G(N_c, q_v=None):
            return get_coulomb_kernel(
                pdi,
                N_c,
                truncation=self.truncation,
                wstc=wstc,
                q_v=q_v)**0.5

        for iw, chi0_GG in enumerate(chi0_wGG):
            if np.allclose(q_c, 0):
                einv_GG = np.zeros((nG, nG), complex)
                for iqf in range(len(gamma_int.qf_qv)):
                    chi0_GG[0, :] = gamma_int.a0_qwG[iqf, iw]
                    chi0_GG[:, 0] = gamma_int.a1_qwG[iqf, iw]
                    chi0_GG[0, 0] = gamma_int.a_wq[iw, iqf]

                    sqrtV_G = get_sqrtV_G(kd.N_c, q_v=gamma_int.qf_qv[iqf])

                    dfc = DielectricFunctionCalculator(
                        sqrtV_G, chi0_GG, mode=fxc_mode, fv_GG=fv)
                    einv_GG += dfc.get_einv_GG() * gamma_int.weight_q[iqf]
            else:
                sqrtV_G = get_sqrtV_G(kd.N_c)

                dfc = DielectricFunctionCalculator(
                    sqrtV_G, chi0_GG, mode=fxc_mode, fv_GG=fv)
                einv_GG = dfc.get_einv_GG()

            if self.ppa:
                einv_wGG.append(einv_GG - delta_GG)
            else:
                W_GG = chi0_GG
                W_GG[:] = (einv_GG - delta_GG) * (sqrtV_G *
                                                  sqrtV_G[:, np.newaxis])

                if self.q0_corrector is not None and np.allclose(q_c, 0):
                    if iw == 0:
                        print_ac = True
                    else:
                        print_ac = False
                    this_w = wblocks1d.a + iw
                    self.add_q0_correction(pdi, W_GG, einv_GG,
                                           chi0_wxvG[this_w],
                                           chi0_wvv[this_w],
                                           sqrtV_G,
                                           print_ac=print_ac)
                elif np.allclose(q_c, 0) or self.integrate_gamma != 0:
                    W_GG[0, 0] = (einv_GG[0, 0] - 1.0) * V0
                    W_GG[0, 1:] = einv_GG[0, 1:] * sqrtV_G[1:] * sqrtV0
                    W_GG[1:, 0] = einv_GG[1:, 0] * sqrtV0 * sqrtV_G[1:]

        if self.ppa:
            omegat_GG = self.E0 * np.sqrt(einv_wGG[1] /
                                          (einv_wGG[0] - einv_wGG[1]))
            R_GG = -0.5 * omegat_GG * einv_wGG[0]
            W_GG = pi * R_GG * sqrtV_G * sqrtV_G[:, np.newaxis]
            if np.allclose(q_c, 0) or self.integrate_gamma != 0:
                W_GG[0, 0] = pi * R_GG[0, 0] * V0
                W_GG[0, 1:] = pi * R_GG[0, 1:] * sqrtV_G[1:] * sqrtV0
                W_GG[1:, 0] = pi * R_GG[1:, 0] * sqrtV0 * sqrtV_G[1:]

            self.timer.stop('Dyson eq.')
            return pdi, blocks1d, [W_GG, omegat_GG]

        # XXX This creates a new, large buffer.  We could perhaps
        # avoid that.  Buffer used to exist but was removed due to #456.
        W_wGG = chi0.blockdist.redistribute(chi0_wGG, chi0.nw)

        self.timer.stop('Dyson eq.')
        return pdi, blocks1d, W_wGG

    @timer('Kohn-Sham XC-contribution')
    def calculate_ks_xc_contribution(self):
        name = self.filename + '.vxc.npy'
        fd, vxc_skn = self.read_contribution(name)
        if vxc_skn is None:
            print('Calculating Kohn-Sham XC contribution', file=self.fd)
            self.fd.flush()
            vxc_skn = vxc(self.gs, self.gs.hamiltonian.xc) / Ha
            n1, n2 = self.bands
            vxc_skn = vxc_skn[:, self.kpts, n1:n2]
            np.save(fd, vxc_skn)
            fd.close()
        return vxc_skn

    @timer('EXX')
    def calculate_exact_exchange(self):
        name = self.filename + '.exx.npy'
        fd, exx_skn = self.read_contribution(name)
        if exx_skn is None:
            print('Calculating EXX contribution', file=self.fd)
            self.fd.flush()
            exx = EXX(self.gs, kpts=self.kpts, bands=self.bands,
                      txt=self.filename + '.exx.txt', timer=self.timer)
            exx.calculate()
            exx_skn = exx.get_eigenvalue_contributions() / Ha
            np.save(fd, exx_skn)
            fd.close()
        return exx_skn

    def read_contribution(self, filename):
        fd = opencew(filename)  # create, exclusive, write
        if fd is not None:
            # File was not there: nothing to read
            return fd, None

        try:
            with open(filename, 'rb') as fd:
                x_skn = np.load(fd)
        except IOError:
            print('Removing broken file:', filename, file=self.fd)
        else:
            print('Read:', filename, file=self.fd)
            if x_skn.shape == self.shape:
                return None, x_skn
            print('Removing bad file (wrong shape of array):', filename,
                  file=self.fd)

        if self.world.rank == 0:
            os.remove(filename)

        return opencew(filename), None

    def print_results(self, results):
        description = ['f:      Occupation numbers',
                       'eps:     KS-eigenvalues [eV]',
                       'vxc:     KS vxc [eV]',
                       'exx:     Exact exchange [eV]',
                       'sigma:   Self-energies [eV]',
                       'dsigma:  Self-energy derivatives',
                       'Z:       Renormalization factors',
                       'qp:      QP-energies [eV]']

        print('\nResults:', file=self.fd)
        for line in description:
            print(line, file=self.fd)

        b1, b2 = self.bands
        names = [line.split(':', 1)[0] for line in description]
        ibzk_kc = self.gs.kd.ibzk_kc
        for s in range(self.gs.nspins):
            for i, ik in enumerate(self.kpts):
                print('\nk-point ' +
                      '{0} ({1}): ({2:.3f}, {3:.3f}, {4:.3f})'.format(
                          i, ik, *ibzk_kc[ik]) + '                ' +
                      self.fxc_mode, file=self.fd)
                print('band' +
                      ''.join('{0:>8}'.format(name) for name in names),
                      file=self.fd)

                def actually_print_results(resultset):
                    for n in range(b2 - b1):
                        print('{0:4}'.format(n + b1) +
                              ''.join('{0:8.3f}'
                                      .format(resultset[name][s, i, n])
                                      for name in names),
                              file=self.fd)

                for fxc_mode in results:
                    print(fxc_mode.rjust(69), file=self.fd)
                    actually_print_results(results[fxc_mode])

        self.timer.write(self.fd)

    @timer('PPA-Sigma')
    def calculate_sigma_ppa(self, n_mG, deps_m, f_m, W, *unused):
        W_GG, omegat_GG = W

        sigma = 0.0
        dsigma = 0.0

        for m in range(len(n_mG)):
            deps_GG = deps_m[m]
            sign_GG = 2 * f_m[m] - 1
            x1_GG = 1 / (deps_GG + omegat_GG - 1j * self.eta)
            x2_GG = 1 / (deps_GG - omegat_GG + 1j * self.eta)
            x3_GG = 1 / (deps_GG + omegat_GG - 1j * self.eta * sign_GG)
            x4_GG = 1 / (deps_GG - omegat_GG - 1j * self.eta * sign_GG)
            x_GG = W_GG * (sign_GG * (x1_GG - x2_GG) + x3_GG + x4_GG)
            dx_GG = W_GG * (sign_GG * (x1_GG**2 - x2_GG**2) +
                            x3_GG**2 + x4_GG**2)
            nW_G = np.dot(n_mG[m], x_GG)
            sigma += np.vdot(n_mG[m], nW_G).real
            nW_G = np.dot(n_mG[m], dx_GG)
            dsigma -= np.vdot(n_mG[m], nW_G).real

        x = 1 / (self.qd.nbzkpts * 2 * pi * self.gs.volume)
        return x * sigma, x * dsigma

    def save_restart_file(self, nQ):
        sigma = self.sigmas[self.fxc_mode]
        sigma_eskn_write = sigma.sigma_eskn.copy()
        dsigma_eskn_write = sigma.dsigma_eskn.copy()
        self.world.sum(sigma_eskn_write)
        self.world.sum(dsigma_eskn_write)
        data = {'last_q': nQ,
                'sigma_eskn': sigma_eskn_write + self.previous_sigma,
                'dsigma_eskn': dsigma_eskn_write + self.previous_dsigma,
                'kpts': self.kpts,
                'bands': self.bands,
                'nbands': self.nbands,
                'ecut_e': self.ecut_e,
                'frequencies': self.frequencies,
                'integrate_gamma': self.integrate_gamma}

        if self.world.rank == 0:
            with open(self.restartfile + '.sigma.pckl', 'wb') as fd:
                pickle.dump(data, fd, 2)

    def load_restart_file(self):
        try:
            with open(self.restartfile + '.sigma.pckl', 'rb') as fd:
                data = pickleload(fd)
        except IOError:
            return False
        else:
            if (data['kpts'] == self.kpts and
                data['bands'] == self.bands and
                data['nbands'] == self.nbands and
                (data['ecut_e'] == self.ecut_e).all and
                data['frequencies']['type'] == self.frequencies['type'] and
                data['frequencies']['domega0'] ==
                self.frequencies['domega0'] and
                data['frequencies']['omega2'] == self.frequencies['omega2'] and
                data['integrate_gamma'] == self.integrate_gamma):
                self.last_q = data['last_q']
                self.previous_sigma = data['sigma_eskn']
                self.previous_dsigma = data['dsigma_eskn']
                return True
            else:
                raise ValueError(
                    'Restart file not compatible with parameters used in '
                    'current calculation. Check kpts, bands, nbands, ecut, '
                    'domega0, omega2, integrate_gamma.')

    def calculate_g0w0_outputs(self, sigma):
        eps_skn, f_skn = self.get_eps_and_occs()
        kwargs = dict(
            fd=self.fd,
            shape=self.shape,
            ecut_e=self.ecut_e,
            eps_skn=eps_skn,
            vxc_skn=self.calculate_ks_xc_contribution(),
            exx_skn=self.calculate_exact_exchange(),
            f_skn=f_skn)

        return G0W0Outputs(sigma_eskn=sigma.sigma_eskn,
                           dsigma_eskn=sigma.dsigma_eskn,
                           **kwargs)

    def add_q0_correction(self, pd, W_GG, einv_GG, chi0_xvG, chi0_vv,
                          sqrtV_G, print_ac=False):
        self.q0_corrector.add_q0_correction(
            pd, W_GG, einv_GG, chi0_xvG, chi0_vv,
            sqrtV_G,
            fd=self.fd if print_ac else None)


def choose_bands(bands, relbands, nvalence, nocc):
    if bands is not None and relbands is not None:
        raise ValueError('Use bands or relbands!')

    if relbands is not None:
        bands = [nvalence // 2 + b for b in relbands]

    if bands is None:
        bands = [0, nocc]

    return bands


class G0W0Kernel:
    def __init__(self, xc, **kwargs):
        self.xc = xc
        self.xcflags = XCFlags(xc)
        self._kwargs = kwargs

    def calculate(self, nG, iq, G2G):
        return calculate_kernel(
            xcflags=self.xcflags,
            nG=nG, iq=iq, cut_G=G2G, **self._kwargs)


class G0W0(G0W0Calculator):
    def __init__(self, calc, filename='gw',
                 ecut=150.0,
                 ecut_extrapolation=False,
                 xc='RPA',
                 ppa=False,
                 E0=Ha,
                 Eg=None,
                 eta=0.1,
                 nbands=None,
                 bands=None,
                 relbands=None,
                 frequencies=None,
                 domega0=None,  # deprecated
                 omega2=None,  # deprecated
                 nblocks=1,
                 nblocksmax=False,
                 kpts=None,
                 world=mpi.world,
                 timer=None,
                 **kwargs):
        frequencies = get_frequencies(frequencies, domega0, omega2)

        gpwfile = calc
        calc, context = calc_and_context(gpwfile, filename + '.txt',
                                         world, timer)
        gs = calc.gs_adapter()

        # Check if nblocks is compatible, adjust if not
        if nblocksmax:
            nblocks = get_max_nblocks(context.world, gpwfile, ecut)

        pair = NoCalculatorPairDensity(gs, nblocks=nblocks, context=context)

        kpts = list(select_kpts(kpts, gs.kd))

        if nbands is None:
            nbands = int(gs.volume * (ecut / Ha)**1.5 * 2**0.5 / 3 / pi**2)
        else:
            if ecut_extrapolation:
                raise RuntimeError(
                    'nbands cannot be supplied with ecut-extrapolation.')

        ecut, ecut_e = choose_ecut_things(ecut, ecut_extrapolation)

        if ppa:
            # use small imaginary frequency to avoid dividing by zero:
            frequencies = [1e-10j, 1j * E0]

            parameters = {'eta': 0,
                          'hilbert': False,
                          'timeordered': False}
        else:
            # frequencies = self.frequencies
            parameters = {'eta': eta,
                          'hilbert': True,
                          'timeordered': True}

        from gpaw.response.chi0 import new_frequency_descriptor
        chi_context = context.with_txt(filename + '.w.txt')
        wd = new_frequency_descriptor(
            gs, nbands, frequencies, fd=chi_context.fd)

        chi0calc = Chi0Calculator(
            wd=wd, pair=pair,
            nbands=nbands,
            ecut=ecut,
            intraband=False,
            context=chi_context,
            **parameters)

        bands = choose_bands(bands, relbands, gs.nvalence, chi0calc.nocc2)

        if Eg is None and xc == 'JGMsx':
            Eg = gs.get_band_gap()

        if Eg is not None:
            Eg /= Ha

        xckernel = G0W0Kernel(xc=xc, ecut=ecut / Ha,
                              gs=gs,
                              ns=gs.nspins,
                              wd=wd,
                              Eg=Eg,
                              timer=context.timer,
                              fd=context.fd)

        super().__init__(filename=filename,
                         chi0calc=chi0calc,
                         ecut_e=ecut_e,
                         xckernel=xckernel,
                         eta=eta,
                         ppa=ppa,
                         E0=E0,
                         nbands=nbands,
                         bands=bands,
                         frequencies=frequencies,
                         context=context,
                         kpts=kpts,
                         **kwargs)



from ase import Atoms
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.qed import RRemission
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.tddft.spectrum import photoabsorption_spectrum

# Sodium dimer chain
d = 1.104  # N2 bondlength
L = 8.0  # N2-N2 distance
atoms = Atoms('Na2',
              positions=[[L, 0, +d / 2], [L, 0, -d / 2]],
              cell=[L, 8.0, 8.0])
atoms.center()

# Ground-state calculation
calc = GPAW(mode='lcao', h=0.3, basis='dzp', xc='LDA',
            setups={'Na': '1'},
            convergence={'density': 1e-12},
            txt='gs_nad.out')
atoms.calc = calc
energy = atoms.get_potential_energy()
calc.write('gs.gpw', mode='all')

# Time-propagation calculation
# Read converged ground-state file
td_calc = LCAOTDDFT('gs.gpw', rremission=RRemission(35.05, [0, 0, 1]))
# Attach any data recording or analysis tools
DipoleMomentWriter(td_calc, 'dm_nad.dat')
# Kick
td_calc.absorption_kick([0.0, 0.0, 1e-5])
# Propagate
td_calc.propagate(20, 5000)
# Please note that those parameter values are quite course
# and should be properly converged for subsequent applications.

# Calculate spectrum with small artificial broadening
photoabsorption_spectrum('dm_nad.dat', 'spec_nad.dat',
                         folding='Gauss', width=0.005,
                         e_min=0.0, e_max=10.0, delta_e=0.005)



from gpaw.wannier.w90 import read_wout_all
from myqueue.workflow import run
import numpy as np


def check():
    with open('GaAs.wout') as fd:
        dct = read_wout_all(fd)
    x, y, z = dct['centers'].sum(axis=0)
    w = dct['spreads'].sum()
    a = 5.68
    assert abs(np.array([x, y, z, w]) - [a, a, a, 4.14]).max() < 0.01

    with open('Fe.wout') as fd:
        dct = read_wout_all(fd)
    xyz = dct['centers'].sum(axis=0)
    w = dct['spreads'].sum()
    print(xyz, w)
    assert abs(xyz).max() < 0.005
    assert abs(w - 14.49) < 0.15


def workflow():
    with run(script='GaAs.py'):
        r1 = run(script='GaAs_wannier.py')
    with run(script='Fe.py', cores=8):
        r2 = run(script='Fe_wannier.py', tmax='1h')
    with r1, r2:
        run(function=check)


if __name__ == '__main__':
    check()



from ase import Atoms
from ase.data.s22 import data
from ase.calculators.vdwcorrection import vdWTkatchenko09prl
from ase.parallel import parprint

from gpaw import GPAW, FermiDirac
from gpaw.cluster import Cluster
from gpaw.analyse.hirshfeld import HirshfeldPartitioning
from gpaw.analyse.vdwradii import vdWradii

h = 0.25
box = 3.0

molecule = 'Adenine-thymine_complex_stack'

Energy = {
    'PBE': [],
    'vdW-DF': [],
    'TS09': []}

for molecule in ['Adenine-thymine_complex_stack']:
    ss = Cluster(Atoms(data[molecule]['symbols'],
                       data[molecule]['positions']))

    # split the structures
    s1 = ss.find_connected(0)
    s2 = ss.find_connected(-1)
    assert len(ss) == len(s1) + len(s2)

    c = GPAW(xc='PBE', h=h, nbands=-6,
             occupations=FermiDirac(width=0.1), txt=None)
    cdf = GPAW(xc='vdW-DF', h=h, nbands=-6, occupations=FermiDirac(width=0.1),
               txt=None)

    for s in [s1, s2, ss]:
        s.calc = c
        s.minimal_box(box, h=h)
        Energy['PBE'].append(s.get_potential_energy())
        cc = vdWTkatchenko09prl(HirshfeldPartitioning(c),
                                vdWradii(s.get_chemical_symbols(), 'PBE'))
        s.calc = cc
        Energy['TS09'].append(s.get_potential_energy())

        s.calc = cdf
        Energy['vdW-DF'].append(s.get_potential_energy())

    parprint('Coupled cluster binding energy',
             -data[molecule]['interaction energy CC'] * 1000, 'meV')
    for xc in ['PBE', 'vdW-DF', 'TS09']:
        ene = Energy[xc]
#        print xc, 'energies', ene
        parprint(xc, 'binding energy',
                 (ene[0] + ene[1] - ene[2]) * 1000, 'meV')



from gpaw import restart
from gpaw.lrtddft import LrTDDFT, photoabsorption_spectrum

atoms, calc = restart('na2_gs_unocc.gpw')

# Calculate the omega matrix
lr = LrTDDFT(calc, xc='LDA', restrict={'jend': 5})
# Use only 5 unoccupied states

# Save the omega matrix
lr.write('Omega_Na2.gz')

# Diagonalize the matrix
lr.diagonalize()

# Analyse 5 lowest excitations
lr.analyse(range(5))
photoabsorption_spectrum(lr, 'Na2_spectrum.dat', e_min=0.0, e_max=10)



"""Simple magnons calculation in a 1D hydrogen chain."""

# General modules
import pytest
import numpy as np

# Script modules
from ase import Atoms
from ase.dft.kpoints import monkhorst_pack

from gpaw import PW, GPAW
from gpaw.mpi import world
from gpaw.test import findpeak
from gpaw.response.tms import TransverseMagneticSusceptibility
from gpaw.response.susceptibility import read_macroscopic_component


@pytest.mark.kspair
@pytest.mark.response
def test_response_afm_hchain_gssALDA(in_tmp_dir):
    # ---------- Inputs ---------- #

    # Part 1: Ground state calculation
    # Define atomic structure
    a = 2.5
    vfactor = 0.8
    mm = 1.
    # Ground state calculator options
    xc = 'LDA'
    kpts = 12
    nbands = 2 * (1 + 0)  # 1s + 0 empty shell bands
    ebands = 2 * 1  # Include also 2s bands for numerical consistency
    pw = 250
    conv = {'bands': nbands}

    # # Part 2: Magnetic response calculation
    q_qc = [[0., 0., 0.],
            [1. / 6., 0., 0.],
            [1. / 3., 0., 0.]]
    fxc = 'ALDA'
    fxc_scaling = [True, None, 'afm']
    rshelmax = -1
    rshewmin = 1e-8
    ecut = 120
    frq_w = np.linspace(-0.6, 0.6, 41)
    eta = 0.24
    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    # ---------- Script ---------- #

    # Part 1: Ground state calculation

    Hatom = Atoms('H',
                  cell=[a, 0, 0],
                  pbc=[1, 0, 0])
    Hatom.center(vacuum=vfactor * a, axis=(1, 2))
    Hchain = Hatom.repeat((2, 1, 1))
    Hchain.set_initial_magnetic_moments([mm, -mm])

    calc = GPAW(xc=xc,
                mode=PW(pw),
                kpts=monkhorst_pack((kpts, 1, 1)),
                nbands=nbands + ebands,
                convergence=conv,
                symmetry={'point_group': True},
                parallel={'domain': 1})

    Hchain.calc = calc
    Hchain.get_potential_energy()

    # Part 2: Magnetic response calculation
    fxckwargs = {'rshelmax': rshelmax,
                 'rshewmin': rshewmin,
                 'fxc_scaling': fxc_scaling}
    tms = TransverseMagneticSusceptibility(calc,
                                           nbands=nbands,
                                           fxc=fxc,
                                           eta=eta,
                                           ecut=ecut,
                                           fxckwargs=fxckwargs,
                                           gammacentered=True,
                                           nblocks=nblocks)
    for q, q_c in enumerate(q_qc):
        tms.get_macroscopic_component('+-', q_c, frq_w,
                                      filename=('h-chain_macro_tms'
                                                + '_q%d.csv' % q))

    tms.write_timer()
    world.barrier()

    # Part 3: Identify magnon peak in finite q scattering function
    w0_w, _, chi0_w = read_macroscopic_component('h-chain_macro_tms_q0.csv')
    w1_w, _, chi1_w = read_macroscopic_component('h-chain_macro_tms_q1.csv')
    w2_w, _, chi2_w = read_macroscopic_component('h-chain_macro_tms_q2.csv')

    wpeak1, Ipeak1 = findpeak(w1_w, -chi1_w.imag / np.pi)
    wpeak2, Ipeak2 = findpeak(w2_w, -chi2_w.imag / np.pi)
    mw1 = calculate_afm_mw(wpeak1, eta) * 1000  # to meV
    mw2 = calculate_afm_mw(wpeak2, eta) * 1000  # to meV

    # Part 4: compare new results to test values
    test_fxcs = 1.04744
    test_mw1 = 285.  # meV
    test_mw2 = 494.  # meV  # Remark that mw2 < 2 * mw1 (linear dispersion)
    test_Ipeak1 = 0.0131  # a.u.
    test_Ipeak2 = 0.0290

    # Test fxc_scaling:
    assert abs(fxc_scaling[1] - test_fxcs) < 0.005

    # Magnon peak at q=1/3 q_X:
    assert abs(mw1 - test_mw1) < 10.

    # Magnon peak at q=2/3 q_X:
    assert abs(mw2 - test_mw2) < 10.

    # Scattering function intensity:
    assert abs(Ipeak1 - test_Ipeak1) < 0.005
    assert abs(Ipeak2 - test_Ipeak2) < 0.005

    # Check that spectrum at q=0 vanishes
    chitol = 1e-3 * np.abs(chi1_w.imag).max()
    assert np.abs(chi0_w.imag).max() < chitol

    # Check that the spectrum is antisymmetric around q=0
    assert np.allclose(w0_w[19::-1] + w0_w[21:], 0.)
    assert np.allclose(chi0_w.imag[19::-1] + chi0_w.imag[21:], 0.,
                       atol=0.01 * chitol)
    assert np.allclose(chi1_w.imag[19::-1] + chi1_w.imag[21:], 0.,
                       atol=chitol)
    assert np.allclose(chi2_w.imag[19::-1] + chi2_w.imag[21:], 0.,
                       atol=chitol)


# ---------- Script functionality ---------- #


def calculate_afm_mw(peak_frequency, eta):
    """Assume lorentzian lineshape to compute the magnon frequency
    from the afm magnon lineshape peak position."""
    return np.sqrt(2 * np.sqrt(peak_frequency**4 + eta**2 * peak_frequency**2)
                   - peak_frequency**2 - eta**2)



import pytest
from gpaw.core import UniformGrid


@pytest.mark.ci
def test_fft_interpolation():
    a = UniformGrid(cell=[1, 1, 1], size=(4, 4, 4)).zeros()
    b = UniformGrid(cell=[1, 1, 1], size=(8, 8, 8)).zeros()
    a.data[2, 2, 2] = 1.0
    a.interpolate(out=b)
    assert (b.data[::2, ::2, ::2] == a.data).all()

    b.fft_restrict(out=a)
    assert a.integrate() == pytest.approx(b.integrate(), abs=1e-12)



import pytest
from ase import Atoms
import numpy as np
from gpaw import GPAW, FermiDirac, Davidson, Mixer, restart
from gpaw.cdft.cdft import CDFT
from gpaw.cdft.cdft_coupling import CouplingParameters
from gpaw.mpi import world


@pytest.mark.skipif(world.size > 1, reason='cdft coupling not parallel')
def test_cdft_restart(in_tmp_dir):
    distance = 2.5
    sys = Atoms('He2', positions=([0., 0., 0.], [0., 0., distance]))
    sys.center(3)
    sys.set_pbc(False)
    sys.set_initial_magnetic_moments([0.5, 0.5])

    calc_b = GPAW(h=0.2,
                  mode='fd',
                  basis='dzp',
                  charge=1,
                  xc='PBE', symmetry='off',
                  occupations=FermiDirac(0., fixmagmom=True),
                  eigensolver=Davidson(3),
                  spinpol=True,
                  nbands=4,
                  mixer=Mixer(beta=0.25, nmaxold=3, weight=100.0),
                  txt='He2+_final_%3.2f.txt' % distance,
                  convergence={'eigenstates': 1.0e-4,
                               'density': 1.0e-1,
                               'energy': 1e-1,
                               'bands': 4})

    cdft_b = CDFT(calc=calc_b,
                  atoms=sys,
                  charge_regions=[[1], [0]],
                  charges=[1, 0],
                  charge_coefs=[27, 0],
                  method='L-BFGS-B',
                  txt='He2+_final_%3.2f.cdft' % distance,
                  minimizer_options={'gtol': 0.1})
    sys.calc = cdft_b
    sys.get_potential_energy()
    sys.calc.calc.write('H2.gpw', mode='all')

    # Restart

    atoms, calc = restart('H2.gpw')

    coupling = CouplingParameters(calc_a=calc, calc_b=calc,
                                  wfs_a='H2.gpw', wfs_b='H2.gpw',
                                  Va=[27, 0], Vb=[27, 0],
                                  charge_regions_A=[[1], [0]],
                                  charge_regions_B=[[1], [0]])
    overlaps = coupling.get_pair_density_matrix(calc, calc)[0]
    for i in [0, 1, 2]:
        assert (np.isclose(np.real(overlaps[0, i, i]), 1.))



import numpy as np
import pytest
from gpaw.core.matrix import Matrix
from gpaw.mpi import broadcast_exception, world


@pytest.mark.parametrize('dtype', [float, complex])
def test_inv(dtype):
    if world.size > 1 and dtype == float:
        # Not implemented
        return

    N = 15
    S0 = Matrix(N, N,
                dist=(world, 1, 1),
                dtype=dtype)

    if world.rank == 0:
        S0.data[:] = np.diag(np.arange(1, N + 1))
        if dtype == float:
            S0.data[-1, 0] = 0.1
        else:
            S0.data[-1, 0] = 0.1j

    S = S0.new(dist=(world, world.size, 1, 2))
    S0.redist(S)

    iS = S.copy()
    iS.inv()

    S.tril2full()
    iS.tril2full()

    A = S.multiply(iS)
    A.redist(S0)

    with broadcast_exception(world):
        if world.rank == 0:
            assert abs(S0.data - np.eye(N)).max() < 1e-14



import os

scalapack = True
hdf5 = False
# ld: /usr/local/phdf5-1.8.5/lib/libhdf5.a(H5.o): relocation R_X86_64_32 against `.rodata.str1.4' can not be used when making a shared object; recompile with -fPIC
compiler = 'icc'

mpi='/opt/mpi/bullxmpi/1.1.16.5'
mkl='/usr/local/Intel_compilers/c/composer_xe_2011_sp1.7.256/mkl/lib/intel64'
intel='/usr/local/Intel_compilers/c/composer_xe_2011_sp1.7.256/compiler/lib/intel64'
hdf='/usr/local/phdf5-1.8.5'
#
# cublasZdgmm does not exist in cuda 4.2
# /tmp/ipo_iccjq2M5h1.o: In function `cudgmm':
# ipo_out1.c:(.text.hot0001d+0x522b): undefined reference to `cublasZdgmm'
# strings /usr/local/cuda-4.2/lib64/libcublas.so | grep "cublasZdgmm"
cuda='/usr/local/cuda-4.2'  # comment out if no cuda

libraries =[
    'cublas', 'cufft', 'cuda',  # comment out if no cuda
    'cudart',  # comment out if no cuda
    #'mkl_def',
    'mkl_scalapack_lp64', 'mkl_intel_lp64', 'mkl_sequential',
    'mkl_core', 'mkl_blacs_openmpi_lp64',
    #'hdf5',
    'mpi',
    ]
library_dirs =[
    intel,
    os.path.join(mpi, 'lib'),
    mkl,
    os.path.join(cuda, 'lib64'),  # comment out if no cuda
    #os.path.join(hdf, 'lib'),
    ]
include_dirs +=[
    os.path.join(mpi, 'include'),
    os.path.join(cuda, 'include'),  # comment out if no cuda
    #os.path.join(hdf, 'include'),
    ]
extra_link_args =[
    '-Wl,-rpath=' + intel +
    ',-rpath=' + os.path.join(mpi, 'lib') +
    ',-rpath=' + os.path.join(cuda, 'lib64') +  # comment out if no cuda
    ',-rpath=' + mkl
    #',-rpath=' + os.path.join(hdf, 'lib')
    ]
extra_compile_args =['-xHOST', '-O3', '-ipo', '-std=c99', '-fPIC', '-Wall']
extra_objects += ['./c/cukernels.o']
define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]
mpicompiler = os.path.join(mpi, 'bin', 'mpicc')
mpilinker = mpicompiler




from ase.build import bulk
from gpaw import GPAW
from gpaw.raman.elph import EPC

supercell = (2, 2, 2)
atoms = bulk('C', 'diamond', a=3.567)
atoms_sc = atoms * supercell
calc = GPAW(mode='lcao', basis='dzp',
            kpts=(5, 5, 5), xc='PBE',
            symmetry={'point_group': False},
            convergence={'bands': 'nao', 'density': 1e-5},
            parallel={'domain': 1},
            txt='gs_super.txt')
atoms_sc.calc = calc
atoms_sc.get_potential_energy()

# Supercell matrix
elph = EPC(atoms, supercell=supercell)
elph.calculate_supercell_matrix(calc, include_pseudo=True)



import numpy as np
from ase import Atoms
from gpaw import GPAW, PW, FermiDirac
from gpaw.mpi import world


def test_pw_par_strategies(in_tmp_dir):
    ecut = 200
    kpoints = [1, 1, 4]
    atoms = Atoms('HLi', cell=[6, 6, 3.4], pbc=True,
                  positions=[[3, 3, 0], [3, 3, 1.6]])

    for xc in ['LDA', 'PBE']:
        def calculate(d, k):
            label = 'gpaw.{xc}.domain{d}.kpt{k}'.format(xc=xc, d=d, k=k)
            atoms.calc = GPAW(mode=PW(ecut),
                              xc=xc,
                              txt=label + '.txt',
                              parallel={'domain': d, 'kpt': k},
                              kpts={'size': kpoints},
                              occupations=FermiDirac(width=0.1))

            def stopcalc():
                atoms.calc.scf.converged = True
            atoms.calc.attach(stopcalc, 4)

            e = atoms.get_potential_energy()
            f = atoms.get_forces()
            s = atoms.get_stress()
            atoms.calc.write(label + '.gpw', mode='all')
            GPAW(label + '.gpw', txt=None)
            return e, f, s

        for d in [1, 2, 4, 8]:
            for k in [1, 2]:
                if d * k > world.size:
                    continue
                e2, f2, s2 = calculate(d, k)
                if d + k == 2:
                    e1, f1, s1 = e2, f2, s2
                else:
                    eerr = abs(e2 - e1)
                    ferr = np.abs(f2 - f1).max()
                    serr = np.abs(s2 - s1).max()
                    if world.rank == 0:
                        print('errs', d, k, eerr, ferr, serr)
                    assert eerr < 1e-11, 'bad {} energy: err={}'.format(
                        xc, eerr)
                    assert ferr < 3e-11, 'bad {} forces: err={}'.format(
                        xc, ferr)
                    assert serr < 1e-11, 'bad {} stress: err={}'.format(
                        xc, serr)



import matplotlib.pyplot as plt

from gpaw import GPAW
from gpaw.zero_field_splitting import zfs, convert_tensor


distance = []
coupling = []
for n in range(3, 9, 2):
    name = f'C{n}H{2 * n}'
    calc = GPAW(name + '.gpw')
    d = calc.atoms.get_distance(0, -3)
    D_vv = zfs(calc)
    D, E, axis, D_vv = convert_tensor(D_vv, 'MHz')
    print(d, D, E, axis)
    print(D_vv)
    distance.append(d)
    coupling.append(D)


plt.plot(distance, coupling)
plt.show()



from ase.build import bulk
from gpaw import GPAW
from gpaw.response.df import DielectricFunction

# Part 1: Ground state calculation
atoms = bulk('Si', 'diamond', a=5.431)
calc = GPAW(mode='pw', kpts=(4, 4, 4))

atoms.calc = calc
atoms.get_potential_energy()  # Ground state calculation is performed
calc.write('si.gpw', 'all')  # Use 'all' option to write wavefunction

# Part 2 : Spectrum calculation
# DF: dielectric function object
# Ground state gpw file (with wavefunction) as input
df = DielectricFunction(calc='si.gpw',
                        domega0=0.05)    # Using nonlinear frequency grid
# By default, a file called 'df.csv' is generated
df.get_dielectric_function()



from myqueue.workflow import run


def workflow():
    with run(script='Be_gs_8bands.py', cores=2, tmax='20m'):
        run(script='Be_8bands_lrtddft.py', cores=2, tmax='20m')
        run(script='Be_8bands_lrtddft_dE.py', cores=2, tmax='20m')
    run(script='Na2_relax_excited.py', cores=4, tmax='8h')



"""Poly-ethylene 1-d chain."""
import pytest
import numpy as np
from ase.neighborlist import neighbor_list

from gpaw import GPAW
from gpaw.wannier import calculate_overlaps


def check(atoms, rcut=0.8):
    """Make sure all Wannier function centers have two neighbors."""
    atoms.wrap()
    i = neighbor_list('i', atoms, rcut)
    nneighbors = np.bincount(i)
    for symbol, nn in zip(atoms.symbols, nneighbors):
        if symbol == 'X':
            assert nn == 2


@pytest.mark.serial
def test_pe_w90(gpw_files, in_tmp_dir):
    calc = GPAW(gpw_files['c2h4_pw_nosym_wfs'])
    o = calculate_overlaps(calc, n2=6, nwannier=6,
                           projections={'C': 's', 'H': 's'})
    try:
        w = o.localize_w90('pe')
    except FileNotFoundError:
        return  # no wannier90.x executable
    check(w.centers_as_atoms())


@pytest.mark.serial
def test_pe_er(gpw_files):
    calc = GPAW(gpw_files['c6h12_pw_wfs'])
    o = calculate_overlaps(calc, n2=3 * 6, nwannier=18)
    w = o.localize_er()
    check(w.centers_as_atoms())



# flake8: noqa
class PAW(ASECalculator):
    def __init__(self, restart=None, **kwargs):
        ASECalculator.__init__(self)
        self.density = Density()
        self.hamiltonian = Hamiltonian()
        self.wfs = WaveFunctions()

        if restart:
            self.read(Reader(restart))
            self.update(self.atoms, **kwargs)

    def read(self, reader):
        self.atoms = reader.read_atoms()
        self.density.read(reader)
        self.hamiltonian.read(reader)
        self.wfs = self.wfs.read(reader)

    def update(self, atoms, **kwargs):
        """Lazy update."""
        self.density.update(self.atoms, kwargs)
        self.hamiltonian.update(self.atoms, kwargs)

        # If we change mode, we could get a completely new type of
        # wave function object:
        self.wfs = self.wfs.update(self.atoms, kwargs)

    def set(self, **kwargs):
        self.update(self,atoms, **kwargs)

    def allocate(self, wfs=False, lfc=True):
        self.density.allocate(lfc)
        self.hamiltonian.allocate(lfc)
        self.wfs.allocate(wfs, lfc)

    def calculate(self, atoms):
        self.update(atoms)
        self.allocate(wfs=True)
        ...

    def get_potential_energy(self, atoms):
        ASECalculator.get_potential_energy(self, atoms)
        return self.hamiltonian.energy




"""
Calculate the magnetic response in iron using ALDA.

Fast test, where the kernel is scaled to fulfill the Goldstone theorem.
"""

# Workflow modules
import pytest
import numpy as np

# Script modules
import time

from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack
from ase.parallel import parprint

from gpaw import GPAW, PW
from gpaw.response.tms import TransverseMagneticSusceptibility
from gpaw.response.susceptibility import read_macroscopic_component
from gpaw.test import findpeak, equal
from gpaw.mpi import world


@pytest.mark.kspair
@pytest.mark.response
def test_response_iron_sf_gssALDA(in_tmp_dir):
    # ------------------- Inputs ------------------- #

    # Part 1: ground state calculation
    xc = 'LDA'
    kpts = 4
    nb = 6
    pw = 300
    a = 2.867
    mm = 2.21

    # Part 2: magnetic response calculation
    q_qc = [[0.0, 0.0, 0.0], [0.0, 0.0, 1. / 4.]]  # Two q-points along G-N
    frq_qw = [np.linspace(-0.080, 0.120, 26), np.linspace(0.100, 0.300, 26)]
    fxc = 'ALDA'
    fxc_scaling = [True, None, 'fm']
    ecut = 300
    eta = 0.01
    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    # ------------------- Script ------------------- #

    # Part 1: ground state calculation

    t1 = time.time()

    Febcc = bulk('Fe', 'bcc', a=a)
    Febcc.set_initial_magnetic_moments([mm])

    calc = GPAW(xc=xc,
                mode=PW(pw),
                kpts=monkhorst_pack((kpts, kpts, kpts)),
                nbands=nb,
                symmetry={'point_group': False},
                parallel={'domain': 1})

    Febcc.calc = calc
    Febcc.get_potential_energy()
    calc.write('Fe', 'all')
    t2 = time.time()

    # Part 2: magnetic response calculation
    fxckwargs = {'rshelmax': None, 'fxc_scaling': fxc_scaling}
    tms = TransverseMagneticSusceptibility('Fe',
                                           fxc=fxc,
                                           eta=eta,
                                           ecut=ecut,
                                           fxckwargs=fxckwargs,
                                           gammacentered=True,
                                           nblocks=nblocks)

    for q in range(2):
        tms.get_macroscopic_component(
            '+-', q_qc[q], frq_qw[q],
            filename='iron_dsus' + '_%d.csv' % (q + 1))
        tms.write_timer()

    t3 = time.time()

    parprint('Ground state calculation took', (t2 - t1) / 60, 'minutes')
    parprint('Excited state calculation took', (t3 - t2) / 60, 'minutes')

    world.barrier()

    # Part 3: identify magnon peaks in scattering function
    w1_w, chiks1_w, chi1_w = read_macroscopic_component('iron_dsus_1.csv')
    w2_w, chiks2_w, chi2_w = read_macroscopic_component('iron_dsus_2.csv')

    print(w1_w, -chi1_w.imag)
    print(w2_w, -chi2_w.imag)

    wpeak1, Ipeak1 = findpeak(w1_w, -chi1_w.imag)
    wpeak2, Ipeak2 = findpeak(w2_w, -chi2_w.imag)

    mw1 = wpeak1 * 1000
    mw2 = wpeak2 * 1000

    # Part 4: compare new results to test values
    test_fxcs = 1.033
    test_mw1 = -0.03  # meV
    test_mw2 = 176.91  # meV
    test_Ipeak1 = 71.20  # a.u.
    test_Ipeak2 = 44.46  # a.u.

    # fxc_scaling:
    equal(fxc_scaling[1], test_fxcs, 0.005)

    # Magnon peak:
    equal(mw1, test_mw1, 0.1)
    equal(mw2, test_mw2, eta * 650)

    # Scattering function intensity:
    equal(Ipeak1, test_Ipeak1, 5)
    equal(Ipeak2, test_Ipeak2, 5)



# web-page: delta-ne-phi.png

import pickle
import numpy as np
from matplotlib import pyplot
import ase.io


def get_potential_trace(filename):
    with open(filename, 'rb') as f:
        phi = pickle.load(f)
    phi = phi.mean(axis=(0, 1))
    return phi


def get_electron_trace(filename):
    with open(filename, 'rb') as f:
        n = pickle.load(f)
    # Multiply each point by its local cell volume to convert from
    # electron density to electrons in each cube.
    vol_adjust = atoms.cell.volume / np.prod(n.shape)
    n *= vol_adjust
    z_trace = np.sum(n, axis=(0, 1))
    z_trace = np.cumsum(z_trace)
    return z_trace


excess_electrons = [-0.2, -0.1, 0., 0.1, 0.2]
atoms = ase.io.read('atoms0.0000.traj')

fig, axes = pyplot.subplots(nrows=2, figsize=(6.4, 4.0), sharex=True)
fig.subplots_adjust(right=0.99, top=0.99)

###########################
# Potential axis.
###########################

ax = axes[0]
z_trace0 = get_potential_trace('esp0.0000.pckl')
z_trace0 -= z_trace0[0]

for excess_electron in excess_electrons:
    label = '{:.4f}'.format(excess_electron)
    z_trace = get_potential_trace(f'esp{label}.pckl')
    z_trace -= z_trace[0]
    diff = z_trace - z_trace0
    zs = np.linspace(0., atoms.cell[2][2], num=len(diff))
    ax.plot(zs, diff, color='C1', linewidth=2)
    ax.text(zs[-1] + 1., diff[-1], '{:+.1f}'.format(excess_electron),
            ha='left', va='center', color='C1')

###########################
# Electrons axis.
###########################

ax = axes[1]
z_trace0 = get_electron_trace('allelectrondensity0.0000.pckl')

for excess_electron in excess_electrons:
    label = '{:.4f}'.format(excess_electron)
    z_trace = get_electron_trace(f'allelectrondensity{label}.pckl')
    diff = z_trace - z_trace0
    ax.plot(zs, diff, color='C0')
    ax.text(zs[-1] + 1., diff[-1], '{:+.1f}'.format(excess_electron),
            ha='left', va='center', color='C0')

###########################
# Touch up.
###########################

axes[0].set_xlim(0., zs[-1] + 4.)
axes[0].set_ylabel(r'$\Delta \phi_{xy}$, V')
axes[1].set_ylabel(r'$\Delta n_{xy}$')
axes[1].set_xlabel(r'$z$ coordinate, $\mathrm{\AA}$')

fig.savefig('delta-ne-phi.png')



"""PDOS tests."""
import pytest
from gpaw import GPAW


def test_li_pdos_pxyz(gpw_files):
    """Test pdos method with and without m."""
    dos = GPAW(gpw_files['bcc_li_pw']).dos()
    energies = dos.get_energies(npoints=100)
    p1 = dos.raw_pdos(energies, a=0, l=1)
    p2 = sum(dos.raw_pdos(energies, a=0, l=1, m=m) for m in range(3))
    assert abs(p1 - p2).max() < 1e-7


def test_pdos_soc(gpw_files):
    """Test pdos with soc."""
    dos = GPAW(gpw_files['c6h12_pw'], txt='-').dos(soc=True)
    energies = [-10.0]
    p = dos.raw_pdos(energies, a=13, l=0)
    assert p[0] == pytest.approx(1.236, abs=0.01)



"""Test gpaw symmetry command."""
import pytest
from gpaw.cli.main import main

result = """\
symmetry:
  number of symmetries: 48
  number of symmetries with translation: 0

bz sampling:
  number of bz points: 512
  number of ibz points: 29
  monkhorst-pack size: [8, 8, 8]
  monkhorst-pack shift: [0.0625, 0.0625, 0.0625]

"""


@pytest.mark.serial
def test_symmetry(gpw_files, capsys):
    args = ['symmetry',
            str(gpw_files['bcc_li_pw']),
            '-k',
            '{density:3,gamma:1}']
    main(args)
    out = capsys.readouterr().out
    print(out)
    assert out == result



import runpy
import sys


class CLICommand:
    """Run GPAW's Python interpreter."""

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--dry-run', '-z', type=int, default=0,
                            metavar='NCPUS',
                            help='Dry run on NCPUS cpus.')
        parser.add_argument('-d', '--debug', action='store_true',
                            help='Run in debug-mode.')
        group = parser.add_mutually_exclusive_group()
        group.add_argument('--command', '-c', dest='cmd',
                           help='Program passed in as string.')
        group.add_argument('--module', '-m',
                           help='Run library module as a script.')
        parser.add_argument('arguments', metavar='ARG',
                            help='Arguments passed to program in '
                            'sys.argv[1:].',
                            nargs='*')

    @staticmethod
    def run(args):
        if args.cmd:
            sys.argv[:] = ['-c'] + args.arguments
            d = {}
            exec(args.cmd, d, d)
        elif args.module:
            sys.argv[:] = [args.module] + args.arguments
            runpy.run_module(args.module, run_name='__main__', alter_sys=True)
        else:
            sys.argv[:] = args.arguments
            runpy.run_path(args.arguments[0], run_name='__main__')



from gpaw import GPAW
from ase.build import molecule

calc = GPAW()
mol = molecule('C6H6', calculator=calc)
mol.center(vacuum=5)
E = mol.get_potential_energy()
nt = calc.get_pseudo_density()
n_ae = calc.get_all_electron_density()
# --- literalinclude division line ---
n_ae_fine = calc.get_all_electron_density(gridrefinement=2)



import numpy as np


class DielectricFunctionCalculator:
    def __init__(self, sqrV_G, chi0_GG, mode, fv_GG=None):
        self.sqrV_G = sqrV_G
        self.chiVV_GG = chi0_GG * sqrV_G * sqrV_G[:, np.newaxis]

        self.I_GG = np.eye(len(sqrV_G))

        self.fv_GG = fv_GG
        self.chi0_GG = chi0_GG
        self.mode = mode

    def _chiVVfv_GG(self):
        assert self.mode != 'GW'
        assert self.fv_GG is not None
        return self.chiVV_GG @ self.fv_GG

    def e_GG_gwp(self):
        gwp_inv_GG = np.linalg.inv(self.I_GG - self._chiVVfv_GG() +
                                   self.chiVV_GG)
        return self.I_GG - gwp_inv_GG @ self.chiVV_GG

    def e_GG_gws(self):
        # Note how the signs are different wrt. gwp.
        # Nobody knows why.
        gws_inv_GG = np.linalg.inv(self.I_GG + self._chiVVfv_GG() -
                                   self.chiVV_GG)
        return gws_inv_GG @ (self.I_GG - self.chiVV_GG)

    def e_GG_plain(self):
        return self.I_GG - self.chiVV_GG

    def e_GG_w_fxc(self):
        return self.I_GG - self._chiVVfv_GG()

    def get_e_GG(self):
        mode = self.mode
        if mode == 'GWP':
            return self.e_GG_gwp()
        elif mode == 'GWS':
            return self.e_GG_gws()
        elif mode == 'GW':
            return self.e_GG_plain()
        elif mode == 'GWG':
            return self.e_GG_w_fxc()
        raise ValueError(f'Unknown mode: {mode}')

    def get_einv_GG(self):
        e_GG = self.get_e_GG()
        return np.linalg.inv(e_GG)



import ase.io
from gpaw import GPAW
atoms = ase.io.read('alpra_Mg.CONTCAR')
atoms.pbc = True
atoms.calc = GPAW(xc='vdW-DF',
                  occupations={'name': 'fermi-dirac', 'width': 0.1})
ncpus = 16



import sys
import matplotlib.pyplot as plt
from gpaw import GPAW

# The following five lines read a file name and an optional width
# from the command line.
filename = sys.argv[1]
if len(sys.argv) > 2:
    width = float(sys.argv[2])
else:
    width = None

calc = GPAW(filename, txt=None)
try:
    ef = calc.get_fermi_level()
except ValueError:
    ef = 0
energy, dos = calc.get_dos(spin=0, width=width)
plt.plot(energy - ef, dos)
if calc.get_number_of_spins() == 2:
    energy, dos = calc.get_dos(spin=1, width=width)
    plt.plot(energy - ef, dos)
    plt.legend(('up', 'down'), loc='upper left')
plt.xlabel(r'$\epsilon - \epsilon_F \ \rm{(eV)}$')
plt.ylabel('Density of States (1/eV)')
plt.show()
